from typing import Type, Union

from .base import SharedData, SharedMemoryWrapper


class BytesSharedData(SharedData):
    def __init__(self, shm_wrapper: SharedMemoryWrapper, size: int, cls_byte: Type[Union[bytes, bytearray]]):
        self._size = size
        self._cls_byte = cls_byte
        super().__init__(shm_wrapper)

    def get_data(self) -> bytes:
        if self.shm_wrapper.buf is None:
            raise ValueError('No shared memory buffer')
        return self._cls_byte(self.shm_wrapper.buf[0:self._size])

    @classmethod
    def create_from_data(cls, data: Union[bytes, bytearray]) -> 'BytesSharedData':
        nbytes = len(data)
        shm_wrapper = SharedMemoryWrapper(nbytes)
        shared_data = cls(shm_wrapper, nbytes, type(data))
        shm_wrapper.buf[0:nbytes] = data
        return shared_data



import logging
import sys

LOGGER_NAME = 'aqueduct'

log = logging.getLogger(LOGGER_NAME)
log.setLevel(logging.DEBUG)
formatter = logging.Formatter('[%(asctime)s] %(levelname)s [Flow] [pid:%(process)d] %(message)s')
ch = logging.StreamHandler(sys.stderr)
ch.setFormatter(formatter)
log.addHandler(ch)


def replace_logger(logger: logging.Logger) -> None:
    """Replaces default aqueduct logger with custom one"""
    log.name = logger.name
    log.level = logger.level
    log.parent = logger.parent
    log.propagate = logger.propagate
    log.handlers = logger.handlers
    log.disabled = logger.disabled



class TestCollector:
    async def test_collect_qsize(self, simple_flow):
        qsize_metrics = simple_flow._metrics_manager.exporter.target.queue_sizes.items
        assert len(qsize_metrics) == 4
        for name, value in qsize_metrics:
            assert value == 0



from dataclasses import asdict, dataclass
from typing import Iterable, Tuple

from . import IMetricsItems


@dataclass
class ProcessesStats(IMetricsItems):
    dead: int = 0
    running: int = 0

    @property
    def items(self) -> Iterable[Tuple[str, int]]:
        return asdict(self).items()

    def extend(self, stats: 'ProcessesStats'):
        self.dead += stats.dead
        self.running += stats.running

    def add_dead_process(self):
        self.dead += 1

    def add_running_process(self):
        self.running += 1



import asyncio
import os
import queue
import time
from typing import List
from unittest.mock import MagicMock, patch

import numpy as np
import pytest

from aqueduct.exceptions import (
    FlowError,
    NotRunningError,
)
from aqueduct.flow import (
    Flow,
    FlowState,
    FlowStep,
)
from aqueduct.handler import BaseTaskHandler
from aqueduct.metrics import MAIN_PROCESS
from aqueduct.task import BaseTask, DEFAULT_PRIORITY
from tests.unit.conftest import (
    Task,
    run_flow,
    terminate_worker,
)
from tests.unit.test_shm import ArrayFieldTask

TASKS_BATCH_SIZE = 16


class HandlerWithError(BaseTaskHandler):
    def handle(self, *tasks: BaseTask):
        raise ValueError


@pytest.fixture
async def flow_with_failed_handler(loop):
    async with run_flow(Flow(HandlerWithError())) as flow:
        yield flow


class CatDetector:
    """GPU model emulator that predicts the presence of the cat in the image."""
    IMAGE_PROCESS_TIME = 0.01
    BATCH_REDUCTION_FACTOR = 0.7
    OVERHEAD_TIME = 0.025
    BATCH_PROCESS_TIME = IMAGE_PROCESS_TIME * TASKS_BATCH_SIZE * BATCH_REDUCTION_FACTOR + OVERHEAD_TIME

    def predict(self, images: np.array) -> np.array:
        """Always says that there is a cat in the image.

        The image is represented by a one-dimensional array.
        The model spends less time for processing batch of images due to GPU optimizations. It's emulated
        with BATCH_REDUCTION_FACTOR coefficient.
        """
        batch_size = images.shape[0]
        if batch_size == 1:
            time.sleep(self.IMAGE_PROCESS_TIME)
        else:
            time.sleep(self.IMAGE_PROCESS_TIME * batch_size * self.BATCH_REDUCTION_FACTOR)
        return np.ones(batch_size, dtype=bool)


class CatDetectorHandler(BaseTaskHandler):
    def handle(self, *tasks: ArrayFieldTask):
        images = np.array([task.array for task in tasks])
        predicts = CatDetector().predict(images)
        for task, predict in zip(tasks, predicts):
            task.result = predict


@pytest.fixture
async def flow_with_not_batch_handler(loop) -> Flow:
    """Actually the handler has batching, but batch size is one."""
    async with run_flow(Flow(CatDetectorHandler())) as flow:
        yield flow


@pytest.fixture
async def flow_with_batch_handler(loop) -> Flow:
    async with run_flow(Flow(
            FlowStep(
                CatDetectorHandler(),
                batch_size=TASKS_BATCH_SIZE,
                batch_timeout=0.5,
            ))) as flow:
        yield flow


@pytest.fixture
async def flow_with_dynamic_batch_handler(loop) -> Flow:
    async with run_flow(Flow(
            FlowStep(
                CatDetectorHandler(),
                batch_size=TASKS_BATCH_SIZE,
            ))) as flow:
        yield flow


async def process_tasks(flow: Flow, tasks: List[ArrayFieldTask]):
    await asyncio.gather(*(flow.process(task) for task in tasks))


@pytest.fixture
def tasks_batch(array):
    return [ArrayFieldTask(array) for _ in range(TASKS_BATCH_SIZE)]


class ShareArrayHandler(BaseTaskHandler):
    def __init__(self, array: np.array):
        super().__init__()
        self._array = array

    def handle(self, *tasks: BaseTask):
        for task in tasks:
            task.array = self._array
            task.share_value('array')


class UpdateSharedArrayFieldHandler(BaseTaskHandler):
    def handle(self, *tasks: BaseTask):
        for task in tasks:
            task.array[0] = task.array[-1]  # noqa


@pytest.fixture
async def flow_for_shared_array(loop, array) -> Flow:
    async with run_flow(Flow(ShareArrayHandler(array), UpdateSharedArrayFieldHandler())) as flow:
        yield flow


class MultiProcBatchHandler(BaseTaskHandler):
    def handle(self, *tasks: Task):
        for task in tasks:
            task.result = os.getpid()


@pytest.fixture
async def flow_with_multiproc_bathcer(loop):
    async with run_flow(
            Flow(
                FlowStep(
                    MultiProcBatchHandler(),
                    nprocs=2,
                    batch_size=TASKS_BATCH_SIZE,
                    batch_timeout=0.5,
                ))
    ) as flow:
        yield flow


class SlowStartingHandler(BaseTaskHandler):
    start_time: int

    def on_start(self):
        time.sleep(self.start_time)

    def handle(self, task: Task):
        task.result = os.getpid()


class TestFlow:
    async def test_start_waiting_for_subproccess_to_start(self, task: Task):
        """flow.start() should block until all subprocesses fully initialized."""
        handler = SlowStartingHandler()
        handler.start_time = 1
        slowest_handler = SlowStartingHandler()
        slowest_handler.start_time = 3
        flow = Flow(FlowStep(handler, nprocs=3), FlowStep(slowest_handler, nprocs=3))

        t0 = time.time()
        flow.start()
        t1 = time.time()

        assert t1 - t0 == pytest.approx(slowest_handler.start_time, 0.5)
        await flow.stop()

    async def test_start_waiting_timeout(self, task: Task):
        """flow.start(timeout) should raise FlowError when starting_timeout expired."""
        handler = SlowStartingHandler()
        handler.start_time = 10
        flow = Flow(FlowStep(handler, nprocs=3))

        timeout = 3
        t0 = time.time()
        with pytest.raises(TimeoutError):
            flow.start(timeout=timeout)
        t1 = time.time()

        assert t1 - t0 == pytest.approx(timeout, 0.5)
        await flow.stop()

    async def test_process_set_result(self, simple_flow: Flow, task: Task):
        assert task.result is None
        result = await simple_flow.process(task)
        assert result
        assert task.result == 'test'

    def test_get_queues_info(self, simple_flow: Flow):
        info = simple_flow._get_queues_info()
        queues = simple_flow._queues
        assert len(info) == len(queues[DEFAULT_PRIORITY])
        assert MAIN_PROCESS in info[queues[DEFAULT_PRIORITY][0].queue]
        assert MAIN_PROCESS in info[queues[DEFAULT_PRIORITY][-1].queue]

    async def test_process_expired_task(self, log_file, aqueduct_logger,
                                        slow_sleep_handlers, slow_simple_flow, task):
        timeouts = [handler._handler_sec for handler in slow_sleep_handlers]

        with pytest.raises(FlowError, match='timeout'):
            await slow_simple_flow.process(task, timeout_sec=timeouts[0])
        await asyncio.sleep(sum(timeouts))
        log_str = log_file.read().decode()

        # task is processed by first handler
        assert f'[{slow_sleep_handlers[0].__class__.__name__}] Have message' in log_str
        # task arrives in the second handler, but it is not processed because it has expired
        assert f'[{slow_sleep_handlers[1].__class__.__name__}] Task expired. Skip' in log_str
        # doesn't arrive in the third handler
        assert f'[{slow_sleep_handlers[2].__class__.__name__}] Have message' not in log_str

    async def test_enqueue_timeout(self, sleep_handlers, simple_flow, task):
        timeouts = [handler._handler_sec for handler in sleep_handlers]
        with patch.object(simple_flow._queues[DEFAULT_PRIORITY][0].queue, 'put', MagicMock(side_effect=queue.Full)):
            with pytest.raises(FlowError, match='timeout'):
                await simple_flow.process(task, timeout_sec=1)
            await asyncio.sleep(sum(timeouts))

    async def test_process_not_running_flow(self, simple_flow, task):
        await simple_flow.stop(graceful=False)
        with pytest.raises(NotRunningError):
            await simple_flow.process(task)

    @pytest.mark.parametrize('priority, expected', [
        (0, ['normal task', 'normal task', 'normal task', 'priority task']),
        (1, ['normal task', 'priority task', 'normal task', 'normal task']),
    ])
    async def test_priority_queue_flow(self, slow_priority_queue_flow, priority, expected):
        results = []
        futures = []
        for _ in range(3):
            task = Task()
            future = asyncio.ensure_future(slow_priority_queue_flow.process(task))
            future.add_done_callback(lambda _: results.append('normal task'))
            futures.append(future)
            await asyncio.sleep(0)
        task = Task()
        task.set_priority(priority)
        future = asyncio.ensure_future(slow_priority_queue_flow.process(task))
        future.add_done_callback(lambda _: results.append('priority task'))
        futures.append(future)

        await asyncio.gather(*futures)

        assert results == expected

    async def test_process_huge_count_of_tasks(self, simple_flow):
        """Count of tasks much more than queue size."""
        # queue size = 20
        n_tasks = 90

        coros = [simple_flow.process(Task()) for _ in range(n_tasks)]
        res = await asyncio.gather(*coros)
        assert len(res) == n_tasks

    async def test_stopped_flow_due_to_terminated_step_process(self, simple_flow: Flow, caplog):
        """Checks that flow and event loop were stopped when child Step process was terminated."""
        assert simple_flow.state == FlowState.RUNNING

        process, handler = await terminate_worker(simple_flow)

        assert simple_flow.state == FlowState.STOPPED
        assert f'The process {process.pid} for {handler.__class__.__name__} handler is dead' in caplog.text

    async def test_stopped_flow_due_to_failed_step_process(self, flow_with_failed_handler, task):
        """Checks that flow and event loop were stopped when child Step process finished with error."""
        assert flow_with_failed_handler.state == FlowState.RUNNING

        with pytest.raises(FlowError):
            await flow_with_failed_handler.process(task)
        await asyncio.sleep(1)

        assert flow_with_failed_handler.state == FlowState.STOPPED

    async def test_stopped_flow_due_to_stop_method(self, simple_flow):
        await simple_flow.stop(graceful=False)
        await asyncio.sleep(0.5)

        assert simple_flow.state == FlowState.STOPPED

    async def test_cancelled_process_reraises_cancelled_error(self, slow_simple_flow, task):
        with pytest.raises(asyncio.CancelledError):
            future = asyncio.ensure_future(slow_simple_flow.process(task))
            await asyncio.sleep(0.1)
            future.cancel()
            await future

    async def test_stop_finish_task_on_graceful(self, sleep_handlers, simple_flow, task):
        """Checks that task will be processed to the end after Flow stop command with graceful mode."""
        t = asyncio.ensure_future(simple_flow.process(task))
        await asyncio.sleep(sleep_handlers[0]._handler_sec)
        assert t.done() is False
        await simple_flow.stop()
        await asyncio.sleep(0.1)

        assert t.result()
        assert simple_flow.state == FlowState.STOPPED

    async def test_stop_abort_task(self, sleep_handlers, simple_flow, task):
        """Checks that task will be aborted after Flow stop command without graceful mode."""
        t = asyncio.ensure_future(simple_flow.process(task))
        await asyncio.sleep(sleep_handlers[0]._handler_sec)
        await simple_flow.stop(graceful=False)
        await asyncio.sleep(0.1)

        assert isinstance(t.exception(), FlowError)
        assert simple_flow.state == FlowState.STOPPED

    async def test_process_performance_without_batching(self, flow_with_not_batch_handler, tasks_batch):
        """Checks that will be received TimeoutError with not batch handler."""
        with pytest.raises(asyncio.TimeoutError):
            await asyncio.wait_for(
                process_tasks(flow_with_not_batch_handler, tasks_batch),
                timeout=CatDetector.BATCH_PROCESS_TIME,
            )

    async def test_process_performance_with_batching(self, flow_with_batch_handler, tasks_batch):
        """Checks that all tasks will be processed on time with batch handler."""
        t0 = time.time()
        await process_tasks(flow_with_batch_handler, tasks_batch),
        t1 = time.time()
        assert pytest.approx(t1 - t0, 0.2) == CatDetector.BATCH_PROCESS_TIME
        assert all(task.result for task in tasks_batch)

    async def test_process_performance_with_dynamic_batching(self, flow_with_dynamic_batch_handler, tasks_batch):
        """Checks that all tasks will be processed on time with dynamic batch handler.
        `Dynamic` turns on when we are not specifying bath_timeout
        """
        await asyncio.wait_for(
            process_tasks(flow_with_dynamic_batch_handler, tasks_batch),
            timeout=CatDetector.BATCH_PROCESS_TIME + CatDetector.IMAGE_PROCESS_TIME + CatDetector.OVERHEAD_TIME,
        )
        assert all(task.result for task in tasks_batch)

    async def test_process_seq_batch_filling(self, tasks_batch, flow_with_multiproc_bathcer):
        """Checks that batches are being filled sequentially."""
        await asyncio.gather(*[flow_with_multiproc_bathcer.process(task) for task in tasks_batch])
        assert len(set(task.result for task in tasks_batch)) == 1

    async def test_process_batch_collecting_metrics(self, flow_with_batch_handler, tasks_batch):
        """Checks that batch collecting time doesn't include the time for waiting for the first element."""
        task = tasks_batch[0]
        await asyncio.sleep(0.1)
        await flow_with_batch_handler.process(task)
        batch_timeout = flow_with_batch_handler._steps[0].batch_timeout
        _, batch_time = flow_with_batch_handler._metrics_manager.collector._metrics.batch_times.items[0]

        assert batch_time == pytest.approx(batch_timeout, rel=1e-1)

    @pytest.mark.parametrize('_type, handlers_not_in_log', [
        (0, ('SleepHandler2',)),
        (1, ('SleepHandler1',)),
        (2, ('SleepHandler1', 'SleepHandler2', 'SetResultSleepHandler')),
    ])
    async def test_handle_condition(
            self,
            log_file,
            aqueduct_logger,
            handle_condition_flow,
            task,
            _type,
            handlers_not_in_log,
    ):
        timeouts = [flow_step.handler._handler_sec for flow_step in handle_condition_flow._steps]
        task._type = _type

        await handle_condition_flow.process(task)
        await asyncio.sleep(sum(timeouts))
        log_str = log_file.read().decode()
        for handler_name in handlers_not_in_log:
            assert f'[{handler_name}] Have message' not in log_str


async def test_segfault_due_to_update_shared_fields(array, array_sh_data, simple_flow):
    t1, t2 = BaseTask(), BaseTask()
    t1.sh_array = array_sh_data
    t2.sh_array = array_sh_data

    await simple_flow.process(t1)
    await simple_flow.process(t2)

    # segfault during access to t2.sh_array due to updated `_shared_fields` field
    assert all(t2.sh_array == array)
    # segfault during access to t1.sh_array due to updated `sh_array` field
    assert all(t1.sh_array == array)


async def test_segfault_due_to_share_field_in_handler(array, flow_for_shared_array):
    t = BaseTask()
    await flow_for_shared_array.process(t)
    assert t.array[0] == array[-1]  # noqa



"""Counts the time of transferring an object to the queue and reading it from it.

Compares different implementations: torch tensor, numpy array, with and without shared memory.
Run: `python -m tests.perf.queue_transfer`
"""
import multiprocessing as mp
import time
from multiprocessing import Queue

import logging
import torch

from aqueduct.shm import SharedFieldsMixin
from .utils import (
    ImViewType,
    MAGIC_NUMBER,
    MPType,
    StopProcess,
    get_image,
    get_mp_classes,
    timeit,
)


class BaseTask:
    def __init__(self, image):
        self.im = image


class Task(SharedFieldsMixin, BaseTask):
    pass


def worker(q_in: Queue):
    while True:
        try:
            start, task = q_in.get()
        except TypeError:
            break
        assert task.im[0][0][0] == MAGIC_NUMBER
        logging.info(f'queue transfer time: {(time.monotonic() - start) * 10**3:.4f}')
        # для правильного замера времени - чтобы gc мог удалить task и вернуть shared память в ОС сразу
        # после обработки задачи, а не после того, как получена новая задача из очереди: q_in.get()
        # todo обсудить момент, что unlink занимает столько же или больше времени, чем трансфер объекта
        # возможно это не так важно, т.к. это время несравнимо со временем работы с объектом
        task = None


def run(view_type: ImViewType, mp_type: MPType, share: bool = False):
    logging.info(f'{view_type.value} image, {mp_type.value} multiprocessing, share: {share}')
    q_class, p_class = get_mp_classes(mp_type)
    task_type = Task if view_type == ImViewType.NUMPY and share else BaseTask

    q_task = q_class()
    p = p_class(target=worker, args=(q_task,))
    p.start()

    for _ in range(10):
        im = get_image()

        if view_type == ImViewType.NUMPY:
            if share:
                with timeit() as t:
                    task = task_type(im)
                    task.share_value('im')
            else:
                with timeit() as t:
                    task = task_type(im)
        elif view_type == ImViewType.TT:
            if share:
                with timeit() as t:
                    tensor = torch.from_numpy(im)
                    np_array = tensor.numpy()
                # it takes less than 0.1 ms
                # logging.info(f'tensor converting time, ms:  {t.seconds * 10 ** 3:.4f}')
                with timeit() as t:
                    task = task_type(tensor)
                    task.im.share_memory_()
            else:
                with timeit() as t:
                    task = task_type(torch.from_numpy(im))
        logging.info(f'task creating time, ms:  {t.seconds * 10**3:.4f}')
        q_task.put((time.monotonic(), task))
        time.sleep(0.001)

    q_task.put(StopProcess())

    p.join()


def main():
    run(ImViewType.NUMPY, MPType.PYTHON)
    run(ImViewType.TT, MPType.TORCH)
    run(ImViewType.TT, MPType.TORCH, True)
    run(ImViewType.NUMPY, MPType.PYTHON, True)


if __name__ == '__main__':
    mp.set_start_method('fork')
    main()



from enum import Enum
from abc import ABC, abstractmethod
from typing import Any, Iterable, List, Tuple

MAIN_PROCESS = 'main_process'

AQUEDUCT = 'aqueduct'


class MetricsTypes(Enum):
    TASK_TIMERS = 'task_timers'
    QUEUE_SIZE = 'queue_size'


class IExtendable(ABC):
    @abstractmethod
    def extend(self, obj):
        """Extends self object by object of the same type."""


class IMetricsItems(IExtendable):
    @property
    @abstractmethod
    def items(self) -> Iterable[Tuple[str, Any]]:
        """Returns metrics as set of name and value pairs."""


class MetricsItems(IMetricsItems):
    """Stores metrics as name and value pairs."""
    def __init__(self, items: List[Tuple[str, Any]] = None):
        self._items: List[Tuple[str, Any]] = items or []

    @property
    def items(self) -> List[Tuple[str, Any]]:
        return self._items

    def add(self, name: str, value: Any):
        self._items.append((name, value))

    def add_items(self, items: List[Tuple[str, Any]]):
        self._items.extend(items)

    def extend(self, metrics: 'MetricsItems'):
        self.add_items(metrics.items)



import asyncio
import multiprocessing
import time
from unittest.mock import AsyncMock

import aiohttp.web
import pytest
from aiohttp import web
from aiohttp.test_utils import TestClient

from aqueduct import Flow
from aqueduct.integrations.aiohttp import (
    AppIntegrator,
    FLOW_NAME,
)
from tests.unit.conftest import (
    SleepHandler,
    terminate_worker,
)


async def init_app(app: web.Application):
    integrator = AppIntegrator(app)
    integrator.add_flow(Flow(SleepHandler(0.001)), with_start=True)

    app.on_shutdown.insert(0, AsyncMock())
    app.on_cleanup.insert(0, AsyncMock())

    return app


def wait_process_stop(proc: multiprocessing.Process, timeout: int):
    stoped = False
    for i in range(10):
        time.sleep(1)
        stoped = not proc.is_alive()
        if stoped:
            break

    assert stoped is True


@pytest.fixture
async def app_with_flow():
    app = web.Application()
    AppIntegrator(app).add_flow(Flow(SleepHandler(0.001)))
    return app


@pytest.fixture
async def app_client(aiohttp_client, app_with_flow) -> TestClient:
    yield await aiohttp_client(app_with_flow)


class TestAppIntegrator:
    def test_app_stops_correctly(self, loop):
        """Flow monitoring should not brake correct application stop.

        Real aiohttp application stops differently then aiohttp_server fixture,
        so to check real life behaviour we need to start AppRunner inside a test."""
        async def ok_exit(_):
            async def stop_task():
                await asyncio.sleep(1)
                raise KeyboardInterrupt

            asyncio.create_task(stop_task())

        app = web.Application()
        app.on_startup.append(ok_exit)

        aiohttp.web.run_app(init_app(app))

        assert app.on_shutdown[0].called is True
        assert app.on_cleanup[0].called is True

    def test_app_exits_on_flow_stop(self):
        def sub_proc():
            async def error_exit(_):
                async def stop_task():
                    await asyncio.sleep(1)
                    await terminate_worker(app[FLOW_NAME])

                asyncio.create_task(stop_task())

            app = web.Application()
            app.on_startup.append(error_exit)

            aiohttp.web.run_app(init_app(app))

            return True

        p = multiprocessing.Process(target=sub_proc)
        p.start()
        time.sleep(1)
        wait_process_stop(p, 10)


    async def test_unittest_teardown(self, app_with_flow: web.Application, app_client):
        """Flow monitoring should not brake unit tests.

        Here we just simulate successful test, there should be not errors on teardown."""
        pass




import multiprocessing as mp
import queue
import signal
import sys
from threading import BrokenBarrierError
from time import monotonic
from typing import Iterator, List, Optional

from .handler import BaseTaskHandler
from .logger import log
from .metrics.timer import (
    Timer,
    timeit,
)
from .queues import FlowStepQueue, select_next_queue
from .task import BaseTask, DEFAULT_PRIORITY, StopTask

MAX_SPIN_ITERATIONS=1000


class Worker:
    """Wrapper over the BaseTaskHandler class.

    Takes a task from the input queue, passes it through task_handler and puts it in the output queue.
    """

    def __init__(
            self,
            queues: List[List[FlowStepQueue]],
            task_handler: BaseTaskHandler,
            batch_size: int,
            batch_timeout: float,
            batch_lock: Optional[mp.RLock],
            read_lock: mp.RLock,
            step_number: int,
    ):
        self._queues = queues
        self._queue_priorities = len(self._queues)
        self.task_handler = task_handler
        self.name = task_handler.__class__.__name__
        self.step_name = self.task_handler.get_step_name(step_number)
        self._batch_size = batch_size
        self._batch_timeout = batch_timeout
        self._batch_lock = batch_lock
        self._stop_task: BaseTask = None  # noqa
        self._read_lock = read_lock
        self._step_number = step_number
        self._read_queues = [
            self._queues[priority][self._step_number - 1].queue
            for priority in reversed(range(self._queue_priorities))
        ]

    def _start(self):
        """Runs something huge (e.g. model) in child process."""
        self.task_handler.on_start()

    def _wait_task_no_timeout(self) -> Optional[BaseTask]:
        # There is a problem with Python MP Queue implementation.
        # queue.get() can raise Empty exception even thou queue is in fact not empty.
        # For example, if we have 10 tasks in a queue, we expect batch size to be 10, but it is not always true, because
        #  after reading, say, 5 tasks, Python queue can tell as that there is nothing left, which is in fact false.
        # In our implementation we use an additional spin over a queue to guarantee consistent results.
        # More info: https://bugs.python.org/issue23582
        task: Optional[BaseTask] = None
        i = 0
        if self._read_lock.acquire(block=False):
            try:
                while not task and i < MAX_SPIN_ITERATIONS:
                    # Limit our spin with maximum of MAX_SPIN_ITERATIONS just in case.
                    # We do not want long, purposeless iteration
                    i += 1

                    for queue_in in self._read_queues:
                        try:
                            task = queue_in.get(block=False)
                        except queue.Empty:
                            continue
                        else:
                            task_size = getattr(queue_in, 'task_size', None)
                            if task_size and not isinstance(task, StopTask):
                                task.metrics.save_task_size(task_size, self.step_name, task.priority)
                            break

                    if task is None:
                        # additionally check queue size to make sure that there is in fact no tasks there.
                        # if size > 0 then Empty exception was fake -> we should try to call get() again
                        # since macos does not support qsize method - ignore that check
                        # (it would be great to make a proper queue later)
                        if sys.platform == 'darwin' or all(queue_in.qsize() == 0 for queue_in in self._read_queues):
                            break
            finally:
                self._read_lock.release()
        return task

    def _wait_task_with_timeout(self, timeout: float) -> Optional[BaseTask]:
        task: Optional[BaseTask] = None
        deadline = monotonic() + timeout
        if self._read_lock.acquire(block=True, timeout=timeout):
            try:
                remaining_timeout = deadline - monotonic()
                if remaining_timeout > 0:
                    mp.connection.wait(
                        [queue._reader for queue in self._read_queues],
                        timeout=remaining_timeout,
                    )
                task = self._wait_task_no_timeout()
            finally:
                self._read_lock.release()
        return task

    def _wait_task(self, timeout: float) -> Optional[BaseTask]:
        task: Optional[BaseTask] = None
        if timeout == 0:
            task = self._wait_task_no_timeout()
        else:
            task = self._wait_task_with_timeout(timeout)

        if not task:
            return

        if isinstance(task, StopTask):
            self._stop_task = task
            return

        log.debug(f'[{self.name}] Have message')

        task.metrics.stop_transfer_timer(self.step_name, task.priority)

        # don't pass an expired task to the next steps
        if task.is_expired():
            log.debug(f'[{self.name}] Task expired. Skip: {task}')
            return

        return task

    def _get_batch_with_timeout(self, batch: List[BaseTask]) -> List[BaseTask]:
        """ Collecting incoming tasks into batch.
        We will be waiting until number of tasks in batch would be equal `batch_size`.
        If batch could not be fully collected in batch_timeout time -> return what we have at that time.
        """
        timeout = self._batch_timeout
        while True:
            with timeit() as passed_time:
                task = self._wait_task(timeout)

            if task:
                batch.append(task)

                if len(batch) == self._batch_size:
                    return batch

            timeout -= passed_time.seconds

            if self._stop_task or timeout <= 0:
                return batch

            # timeout should not be less then 1ms, to avoid unnecessary short sleeps
            timeout = max(timeout, 0.001)

    def _get_batch_dynamic(self, batch: List[BaseTask]) -> List[BaseTask]:
        """ Collecting incoming tasks into batch.
        This method will not be waiting for batch_timeout to collect full batch,
        we just simply get all tasks that are currently in the queue and making batch only from them.
        """
        while True:
            task = None
            task = self._wait_task(timeout=0.0)

            if task:
                batch.append(task)
                if len(batch) == self._batch_size:
                    break
            else:
                break

        return batch

    def _wait_batch(self) -> List[BaseTask]:
        batch = []
        timer = Timer()

        # wait for the first task
        while True:
            # wait for some long amount of time (10 secs), and wait again,
            # until we eventually get a task
            task = self._wait_task(10.0)
            if task:
                batch.append(task)
                timer.start()
                break
            elif self._stop_task:
                return []

        # waiting for the rest of the batch if batch_size > 1
        if self._batch_size > 1:
            if self._batch_timeout > 0:
                batch = self._get_batch_with_timeout(batch)
            else:
                batch = self._get_batch_dynamic(batch)

        timer.stop()
        if self._batch_size > 1:
            batch[0].metrics.batch_times.add(self.step_name, timer.seconds)
            batch[0].metrics.batch_sizes.add(self.step_name, len(batch))

        return batch

    def _iter_batches(self) -> Iterator[Optional[List[BaseTask]]]:
        """ Returns iterator over input task batches.
        If there is no tasks in queue -> block until task appears.
        This iterator takes into account batch_timeout and batch size.
        If input task is expired or filtered by condition - it would not be placed in batch.
        """
        while True:
            if self._batch_lock:
                # to take a queue_in lock for the duration of batch filling time
                with self._batch_lock:
                    tasks_batch = self._wait_batch()
            else:
                tasks_batch = self._wait_batch()

            if tasks_batch:
                yield tasks_batch

            if self._stop_task:
                break

    def _post_handle(self, task: BaseTask):
        task.metrics.start_transfer_timer(self.step_name)
        queue_out = select_next_queue(
            queues=self._queues,
            task=task,
            start_index=self._step_number,
        )
        queue_out.put(task)

    def _fix_signals(self):
        """ Sometimes some web frameworks (e.g. unicorn) override signal handlers.
        If processes use start method fork,
        so when we call flow.stop() processes do not terminate and hang.
        Here we restore the signals we need.
        """
        signal.signal(signal.SIGTERM, signal.Handlers.SIG_DFL)
        signal.signal(signal.SIGINT, signal.default_int_handler)

    def loop(self, pid: int, start_barrier: mp.Barrier):
        """Main worker loop.

        The code below is executed in a new process.
        """
        log.info(f'[Worker] initialising handler {self.name}')
        self._fix_signals()
        self._start()
        log.info(f'[Worker] handler {self.name} ok, waiting for others to start')

        try:
            start_barrier.wait()
        except BrokenBarrierError:
            raise TimeoutError('Starting timeout expired')

        log.info(f'[Worker] handler {self.name} ok, starting loop')

        for tasks_batch in self._iter_batches():
            with timeit() as timer:
                self.task_handler.handle(*tasks_batch)

            for task in tasks_batch:
                task.metrics.handle_times.add(self.step_name, timer.seconds)
                self._post_handle(task)

        if self._stop_task:
            self._queues[DEFAULT_PRIORITY][self._step_number].queue.put(self._stop_task)



import multiprocessing
import time
from threading import Thread
from typing import List, Tuple

import pytest

from aqueduct import BaseTask, BaseTaskHandler
from aqueduct.worker import Worker
from aqueduct.queues import FlowStepQueue


class FakeTaskHandler(BaseTaskHandler):
    def handle(self, *tasks: BaseTask):
        pass


class Task(BaseTask):
    def __init__(self, data):
        super().__init__()
        self.metrics.start_transfer_timer('test')
        self.data = data
        self.set_timeout(5000.)

    def __eq__(self, other: 'Task'):
        return self.data == other.data


class Batch:
    def __init__(self, *data: int):
        self.tasks = [Task(d) for d in data]

    def __eq__(self, other: List[Task]):
        return self.tasks == other


def long_elements_producer(iq: multiprocessing.Queue):
    """Long non-blocking producer of elements."""
    tic_elements = {1: 1, 8: 2, 10: 3, 24: 4}

    for tic in range(25):
        if tic in tic_elements:
            iq.put(Task(tic_elements[tic]))

        time.sleep(0.01)


@pytest.fixture
def worker():
    def init(batch_size, batch_timeout) -> Tuple[Worker, multiprocessing.Queue]:
        iq = multiprocessing.Queue(10000)
        oq = multiprocessing.Queue(10000)
        queues = [
            [
                FlowStepQueue(iq, lambda _: True),
                FlowStepQueue(oq, lambda _: True),
            ]
        ]
        worker = Worker(
            queues=queues,
            task_handler=FakeTaskHandler(),
            batch_size=batch_size,
            batch_timeout=batch_timeout,
            batch_lock=None,
            read_lock=multiprocessing.RLock(),
            step_number=1)
        return worker, iq

    yield init


class TestWorker:
    @pytest.mark.parametrize('elements, batch_size, result', [
        (Batch(1, 2, 3), 1, [Batch(1), Batch(2), Batch(3)]),
        (Batch(1, 2, 3), 2, [Batch(1, 2), Batch(3)]),
        (Batch(1, 2, 3), 3, [Batch(1, 2, 3)]),
        (Batch(1, 2, 3), 4, [Batch(1, 2, 3)]),
    ])
    def test_iter_batches(self, worker, elements, batch_size, result):
        w, iq = worker(batch_size, 1.0)
        for item in elements.tasks:
            iq.put(item)

        it = w._iter_batches()
        for res_batch in result:
            assert next(it) == res_batch

    def test_iter_batches_read_all_queue(self, worker):
        """Tests that batcher read all available tasks from queue all the time.

        Problem is queue.get() can raise Empty exception even thou queue is in fact not empty.
         For example, if we have 10 tasks in a queue, we expect batch size to be 10, but it is not always true, because
         after reading, say, 5 tasks, Python queue can tell us that there is nothing left, which is in fact false.
        In our implementation we use an additional spin over a queue to guarantee consistent results.
        More info: https://bugs.python.org/issue23582"""
        batch_size = 100
        elements = Batch(*range(0, 5000))
        expected = [Batch(*range(i, i + batch_size)) for i in range(0, 5000, batch_size)]

        w, iq = worker(100, 0.)

        for item in elements.tasks:
            iq.put(item)

        it = w._iter_batches()
        for res_batch in expected:
            assert next(it) == res_batch

    @pytest.mark.parametrize('timeout, result', [
        (0.01, [Batch(1), Batch(2), Batch(3), Batch(4)]),
        (0.03, [Batch(1), Batch(2, 3), Batch(4)]),
    ])
    def test_iter_batches_timeouts(self, worker, timeout, result):
        w, iq = worker(batch_size=4, batch_timeout=timeout)

        th = Thread(target=long_elements_producer, args=(iq,))
        th.start()

        it = w._iter_batches()
        for res_batch in result:
            assert next(it) == res_batch

        th.join()



import asyncio
import multiprocessing as mp
from typing import Dict, List

from .base import MetricsTypes
from .collect import Collector
from .export import DummyExporter, Exporter


class MetricsManager:
    def __init__(self, collector: Collector, exporter: Exporter):
        self.collector = collector
        self.exporter = exporter
        self._tasks: List[asyncio.Future] = []

    def start(self, queues_info: Dict[mp.Queue, str]):
        if queues_info and self.collector.is_collectible(MetricsTypes.QUEUE_SIZE):
            self._tasks.append(asyncio.ensure_future(self.collector.collect_qsize(queues_info)))
        self._tasks.append(asyncio.ensure_future(self._export_metrics()))

    def stop(self):
        # Остановка фоновых корутин
        for task in self._tasks:
            task.cancel()
        # Send last metrics
        self.exporter.export(self.collector.extract_metrics())

    async def _export_metrics(self):
        while True:
            self.exporter.export(self.collector.extract_metrics())
            await asyncio.sleep(self.exporter.export_period)


def get_metrics_manager(collector: Collector = None, exporter: Exporter = None) -> MetricsManager:
    if collector is None:
        collector = Collector()
    if exporter is None:
        exporter = DummyExporter()
    return MetricsManager(collector, exporter)



import cffi

ffibuilder = cffi.FFI()

ffibuilder.cdef("""
uint32_t load_uint32(uint32_t *v);
void store_uint32(uint32_t *v, uint32_t n);
uint32_t add_and_fetch_uint32(uint32_t *v, uint32_t i);
uint32_t sub_and_fetch_uint32(uint32_t *v, uint32_t i);
""")

ffibuilder.set_source("aqueduct.shm._atomic", r"""
uint32_t load_uint32(uint32_t *v) {
    return __atomic_load_n(v, __ATOMIC_SEQ_CST);
};
void store_uint32(uint32_t *v, uint32_t n) {
    uint32_t i = n;
    __atomic_store(v, &i, __ATOMIC_SEQ_CST);
};
uint32_t add_and_fetch_uint32(uint32_t *v, uint32_t i) {
    return __atomic_add_fetch(v, i, __ATOMIC_SEQ_CST);
};
uint32_t sub_and_fetch_uint32(uint32_t *v, uint32_t i) {
    return __atomic_sub_fetch(v, i, __ATOMIC_SEQ_CST);
};
""")   # <=

if __name__ == "__main__":
    ffibuilder.compile(verbose=True)



from aqueduct.metrics import MAIN_PROCESS
from aqueduct.metrics.timer import timeit
from aqueduct.task import BaseTask


def test_metrics_count(task: BaseTask):

    # кладем в очередь
    task.metrics.start_transfer_timer(MAIN_PROCESS)
    # достаем из очереди
    task.metrics.stop_transfer_timer('handler1')
    # выполняем полезную работу
    with timeit() as timer:
        pass
    task.metrics.handle_times.add('handler1', timer.seconds)
    # кладем в очередь
    task.metrics.start_transfer_timer('handler1')
    # достаем из очереди
    task.metrics.stop_transfer_timer('handler2')
    # выполняем полезную работу
    task.metrics.handle_times.add('handler2', timer.seconds)
    # кладем в очередь
    task.metrics.start_transfer_timer('handler2')
    # достаем из очереди
    task.metrics.stop_transfer_timer('handler3')
    # выполняем полезную работу
    task.metrics.handle_times.add('handler3', timer.seconds)
    # кладем в очередь
    task.metrics.start_transfer_timer('handler3')
    # достаем из очереди
    task.metrics.stop_transfer_timer('parent_process')
    # total
    task.metrics.handle_times.add('total', timer.seconds)

    assert len(task.metrics.handle_times.items) == 4
    assert len(task.metrics.transfer_times.items) == 4



from typing import Callable
from .task import BaseTask


HandleConditionType = Callable[[BaseTask], bool]


class BaseTaskHandler:
    """The base class for your handlers.
       It is used for logic that does not require a model and special initialization.
       If you need a model that starts for a long time and takes up memory, then use  ModelTaskHandler()."""

    @classmethod
    def get_step_name(cls, step_number: int) -> str:
        return f'step{step_number}_{cls.__name__}'

    def on_start(self):
        """Called at startup in a child process.
        So that the model takes up memory only in the subprocess, not in the parent."""
        pass

    def handle(self, *tasks: BaseTask):
        """Called when a task is received from the queue."""
        raise NotImplementedError



import asyncio
import logging
import os
import signal
from typing import Optional

from aqueduct.flow import Flow
from aqueduct.logger import log, replace_logger

from aiohttp import web


AQUEDUCT_FLOW_NAMES = 'aqueduct_flow_names'
FLOW_NAME = 'flow'
FLOWS_OBSERVER = 'aqueduct_flows_observer'
FLOW_START_TIMEOUT_NAME = 'FLOW_START_TIMEOUT'
FLOW_START_TIMEOUT_DEFAULT = 300.0


async def observe_flows(app: web.Application, check_interval: float = 1.):
    flows = {name: app[name] for name in app[AQUEDUCT_FLOW_NAMES]}
    while True:
        for flow_name, flow in flows.items():
            if not flow.is_running:
                log.info(f'Flow {flow_name} is not running, application will be stopped')
                pid = os.getpid()
                # kill process with SIGTERM to ensure, that stopping would not be delayed by other code (like aiohttp)
                os.kill(pid, signal.SIGTERM)
                return

        await asyncio.sleep(check_interval)


async def run_flows_observer(app):
    app[FLOWS_OBSERVER] = asyncio.create_task(observe_flows(app))


async def stop_flows_observer(app):
    app[FLOWS_OBSERVER].cancel()


async def stop_flows(app):
    flows = [app[name] for name in app[AQUEDUCT_FLOW_NAMES]]
    for flow in flows:  # type: Flow
        await flow.stop()


def get_start_timeout() -> Optional[float]:
    try:
        start_timeout = float(os.environ.get(
            FLOW_START_TIMEOUT_NAME,
            FLOW_START_TIMEOUT_DEFAULT,
        ))
    except ValueError:
        start_timeout = FLOW_START_TIMEOUT_DEFAULT
    if start_timeout <= 0:
        return None
    return start_timeout


class AppIntegrator:
    """Adds to app flows and actions to manage flows and app itself."""
    def __init__(
            self,
            app: web.Application,
            exit_on_fail: bool = True,
            logger: Optional[logging.Logger] = None,
    ) -> None:
        if AQUEDUCT_FLOW_NAMES in app:
            raise RuntimeError('AppIntegrator can be created only once. Reuse existing AppIntegrator.')
        self._app = app
        self._app[AQUEDUCT_FLOW_NAMES] = []
        if exit_on_fail:
            self._app.on_startup.append(run_flows_observer)
            self._app.on_shutdown.append(stop_flows_observer)
        self._app.on_shutdown.append(stop_flows)
        if logger is not None:
            replace_logger(logger)

    def add_flow(self, flow: Flow, flow_name: str = FLOW_NAME, with_start: bool = True):
        if flow_name in self._app:
            raise RuntimeError(f'Flow with name "{flow_name}" already exists in app')
        if with_start:
            flow.start(timeout=get_start_timeout())
        self._app[AQUEDUCT_FLOW_NAMES].append(flow_name)
        self._app[flow_name] = flow



import asyncio
from collections import deque
from typing import (
    List,
    Tuple,
    Union,
)

import pytest

from aqueduct.flow import (
    Flow,
    FlowStep,
)
from aqueduct.metrics.collect import AqueductMetricsStorage
from aqueduct.metrics.export import (
    AQUEDUCT,
    BATCH_SIZE_PREFIX,
    BATCH_TIME_PREFIX,
    DummyExporter,
    HANDLE_TIME_PREFIX,
    QSIZE_PREFIX,
    TASKS_PREFIX,
    TRANSFER_TIME_PREFIX,
    TASK_SIZE_PREFIX,
    ToStatsDMetricsExporter,
)
from tests.unit.conftest import (
    SleepHandler,
    Task,
    run_flow,
)

EXPORT_PERIOD = 0.001


class StatsDMetricsBuffer:
    def __init__(self):
        self._data = deque()
        self._size = 0

    @property
    def data(self) -> List[bytes]:
        return list(self._data)

    def _format(self, name: Union[str, bytes], value: Union[int, float], _type: Union[str, bytes]) -> bytes:
        name = name if isinstance(name, bytes) else name.encode('utf8')
        _type = _type if isinstance(_type, bytes) else _type.encode('utf8')
        return b'%s:%d|%s' % (name, int(value), _type)

    def _set_metric(self, name: Union[str, bytes], value: Union[int, float], _type: Union[str, bytes]):
        data = self._format(name=name, value=value, _type=_type)
        self._data.append(data)
        self._size += len(data)

    def timing(self, name, seconds: Union[float, int]):
        self._set_metric(name, seconds, b'ms')

    def count(self, name, value):
        self._set_metric(name, value, b'c')

    def gauge(self, name, value):
        self._set_metric(name, value, b'g')


@pytest.fixture
def statsd_metrics_buffer() -> StatsDMetricsBuffer:
    return StatsDMetricsBuffer()


@pytest.fixture
def to_statsd_metrics_exporter(statsd_metrics_buffer) -> ToStatsDMetricsExporter:
    return ToStatsDMetricsExporter(statsd_metrics_buffer, export_period=EXPORT_PERIOD)


@pytest.fixture
async def flow_with_exporter(loop, sleep_handlers: Tuple[SleepHandler, ...],
                             to_statsd_metrics_exporter) -> Flow:
    async with run_flow(Flow(*sleep_handlers, metrics_exporter=to_statsd_metrics_exporter)) as flow:
        yield flow


@pytest.fixture
def dummy_metrics_exporter() -> DummyExporter:
    return DummyExporter(export_period=EXPORT_PERIOD)


@pytest.fixture
async def flow_without_metrics(loop, sleep_handlers: Tuple[SleepHandler, ...]) -> Flow:
    async with run_flow(Flow(*sleep_handlers, metrics_enabled=False)) as flow:
        yield flow


@pytest.fixture
async def flow_with_batching_to_statsd(loop, to_statsd_metrics_exporter) -> Flow:
    async with run_flow(
            Flow(
                FlowStep(SleepHandler(0.001), batch_size=4, batch_timeout=0.05),
                metrics_exporter=to_statsd_metrics_exporter,
            )) as flow:
        yield flow


@pytest.fixture
async def flow_with_batching(loop, dummy_metrics_exporter) -> Flow:
    async with run_flow(
            Flow(
                FlowStep(SleepHandler(0.001), batch_size=4, batch_timeout=0.05),
                metrics_exporter=dummy_metrics_exporter,
            )) as flow:
        yield flow


@pytest.fixture
async def tasks() -> List[Task]:
    return [Task() for _ in range(10)]


class TestToAvioMetricsExporter:
    async def test_export(
            self,
            sleep_handlers,
            statsd_metrics_buffer,
            flow_with_exporter: Flow,
            task,
    ):
        await flow_with_exporter.process(task)
        # wait for metrics export
        await asyncio.sleep(2 * EXPORT_PERIOD)

        metrics = [m.decode() for m in statsd_metrics_buffer.data]
        assert sum(1 for m in metrics if m.startswith(f'{AQUEDUCT}.{HANDLE_TIME_PREFIX}')) == 4
        assert sum(1 for m in metrics if m.startswith(f'{AQUEDUCT}.{TRANSFER_TIME_PREFIX}')) == 4
        assert sum(1 for m in metrics if m.startswith(f'{AQUEDUCT}.{TASK_SIZE_PREFIX}')) == 4
        assert sum(1 for m in metrics if m.startswith(f'{AQUEDUCT}.{BATCH_TIME_PREFIX}')) == 0
        assert sum(1 for m in metrics if m.startswith(f'{AQUEDUCT}.{BATCH_SIZE_PREFIX}')) == 0
        assert sum(1 for m in metrics if m.startswith(f'{AQUEDUCT}.{QSIZE_PREFIX}')) == 4
        assert sum(1 for m in metrics if m.startswith(f'{AQUEDUCT}.{TASKS_PREFIX}')) == 1

    async def test_export_batch_metrics_for_avio(
            self,
            flow_with_batching_to_statsd,
            tasks: List[Task],
            statsd_metrics_buffer,
    ):
        await asyncio.gather(*[flow_with_batching_to_statsd.process(task) for task in tasks])
        await asyncio.sleep(2 * EXPORT_PERIOD)

        metrics = [m.decode() for m in statsd_metrics_buffer.data]
        assert sum(1 for m in metrics if m.startswith(f'{AQUEDUCT}.{BATCH_TIME_PREFIX}')) == 3
        assert sum(1 for m in metrics if m.startswith(f'{AQUEDUCT}.{BATCH_SIZE_PREFIX}')) == 3

    async def test_export_batch_metrics(self, dummy_metrics_exporter, flow_with_batching: Flow, tasks: List[Task]):
        await asyncio.gather(*[flow_with_batching.process(task) for task in tasks])
        await asyncio.sleep(2 * EXPORT_PERIOD)

        metrics = dummy_metrics_exporter.target
        t1, t2, t3 = (seconds for name, seconds in metrics.batch_times.items)
        assert t1 == pytest.approx(0.01, 1) and t2 == pytest.approx(0.01, 1) and t3 >= 0.05
        s1, s2, s3 = (sizes for name, sizes in metrics.batch_sizes.items)
        assert s1 == s2 == 4 and s3 == 2

    async def test_export_disabled_metrics(self, sleep_handlers, flow_without_metrics: Flow, task):
        await flow_without_metrics.process(task)
        # wait for metrics export
        await asyncio.sleep(2 * EXPORT_PERIOD)

        metrics: AqueductMetricsStorage = flow_without_metrics._metrics_manager.exporter.target
        assert not metrics.queue_sizes.items
        assert not metrics.transfer_times.items
        assert not metrics.task_sizes.items
        assert not metrics.handle_times.items
        assert not metrics.batch_times.items
        assert not metrics.batch_sizes.items
        assert not metrics.tasks_stats.complete



from typing import Dict
from unittest.mock import Mock

from tests.unit.conftest import Task
from aqueduct.shm.base import SharedData, SharedMemoryWrapper
from aqueduct.task import BaseTask


class IntSharedData(SharedData):
    def __init__(self, shm_wrapper: SharedMemoryWrapper, value: int):
        super().__init__(shm_wrapper)
        self.value = value

    def get_data(self) -> int:
        return self.value

    @classmethod
    def create_from_data(cls, data: int) -> 'IntSharedData':
        return cls(Mock(), data)


def get_task_with_shf(shf_data: Dict[str, SharedData]):
    t = BaseTask()
    for k, v in shf_data.items():
        setattr(t, k, v)
    return t


class TestBaseTask:
    def test_get_queues_info(self, task: Task):
        task.result = 'some result'
        task2 = Task()
        task2.update(task)

        assert task2.task_id != task.task_id, 'Id must be different'
        assert task2.result == task.result, 'Result must be same'
        assert task2.metrics is not task.metrics

    def test_update_shared_fields_source_task_with_shf(self):
        target_task = BaseTask()
        source_task = get_task_with_shf({
            'foo': IntSharedData.create_from_data(1),
            'bar': IntSharedData.create_from_data(2),
        })
        target_task.update(source_task)

        assert target_task._shared_fields == source_task._shared_fields
        assert getattr(target_task, 'foo') == getattr(source_task, 'foo')
        assert getattr(target_task, 'bar') == getattr(source_task, 'bar')

    def test_update_shared_fields_both_tasks_with_shf(self):
        target_task = get_task_with_shf({
            'foo': IntSharedData.create_from_data(1),
            'bar': IntSharedData.create_from_data(2),
        })
        source_task = get_task_with_shf({
            'foo': IntSharedData.create_from_data(3),
            'baz': IntSharedData.create_from_data(4),
        })
        old_tt_shf_info = target_task._shared_fields
        st_shf_info = source_task._shared_fields
        target_task.update(source_task)

        new_tt_shf_info = target_task._shared_fields

        assert new_tt_shf_info['foo'] == old_tt_shf_info['foo']
        assert new_tt_shf_info['bar'] == old_tt_shf_info['bar']
        assert new_tt_shf_info['baz'] == st_shf_info['baz']

        assert getattr(target_task, 'foo') == 1
        assert getattr(target_task, 'bar') == 2
        assert getattr(target_task, 'baz') == 4



import asyncio
import logging
import multiprocessing as mp
import sys
import tempfile
import time
from contextlib import asynccontextmanager
from multiprocessing import (
    Value,
    queues,
)
from typing import Tuple

import numpy as np
import pytest

from aqueduct.flow import Flow, FlowStep
from aqueduct.handler import BaseTaskHandler
from aqueduct.logger import LOGGER_NAME
from aqueduct.multiprocessing import ProcessContext
from aqueduct.shm.numpy import NPArraySharedData
from aqueduct.task import BaseTask


# it's necessary for correct work on MacOS
mp.set_start_method('fork')


class SharedCounter:
    """ A synchronized shared counter.

    The locking done by multiprocessing.Value ensures that only a single
    process or thread may read or write the in-memory ctypes object. However,
    in order to do n += 1, Python performs a read followed by a write, so a
    second process may read the old value before the new one is written by the
    first process. The solution is to use a multiprocessing.Lock to guarantee
    the atomicity of the modifications to Value.

    This class comes almost entirely from Eli Bendersky's blog:
    http://eli.thegreenplace.net/2012/01/04/shared-counter-with-pythons-multiprocessing/

    """

    def __init__(self, n: int = 0):
        self.count = Value('i', n)

    def increment(self, n: int = 1):
        """ Increment the counter by n (default = 1) """
        with self.count.get_lock():
            self.count.value += n

    @property
    def value(self):
        """ Return the value of the counter """
        return self.count.value


class MultiPlatformQueue(queues.Queue):
    """ A portable implementation of multiprocessing.Queue.

    Because of multithreading / multiprocessing semantics, Queue.qsize() may
    raise the NotImplementedError exception on Unix platforms like Mac OS X
    where sem_getvalue() is not implemented. This subclass addresses this
    problem by using a synchronized shared counter (initialized to zero) and
    increasing / decreasing its value every time the put() and get() methods
    are called, respectively. This not only prevents NotImplementedError from
    being raised, but also allows us to implement a reliable version of both
    qsize() and empty().

    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.size = SharedCounter(0)

    def put(self, *args, **kwargs):
        super().put(*args, **kwargs)
        self.size.increment(1)

    def get(self, *args, **kwargs):
        elem = super().get(*args, **kwargs)
        self.size.increment(-1)
        return elem

    def qsize(self):
        """ Reliable implementation of multiprocessing.Queue.qsize() """
        return self.size.value

    def empty(self):
        """ Reliable implementation of multiprocessing.Queue.empty() """
        return not self.qsize()


if sys.platform == 'darwin':
    queues.Queue = MultiPlatformQueue


async def terminate_worker(flow: Flow) -> Tuple[mp.process.BaseProcess, BaseTaskHandler]:
    handler, context = next(iter(flow._contexts.items()))  # type: BaseTaskHandler, ProcessContext
    process = context.processes[0]
    process.terminate()
    await asyncio.sleep(1)
    return process, handler


@pytest.fixture
def array() -> np.ndarray:
    return np.array([1, 2])


@pytest.fixture
def array_sh_data(array: np.ndarray) -> NPArraySharedData:
    return NPArraySharedData.create_from_data(array)


class Task(BaseTask):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.result = None


@pytest.fixture
def task() -> Task:
    return Task()


class SleepHandler(BaseTaskHandler):
    def __init__(self, handler_sec: float):
        self._handler_sec: float = handler_sec

    def handle(self, *tasks: BaseTask):
        time.sleep(self._handler_sec)


class SleepHandler1(SleepHandler):
    pass


class SleepHandler2(SleepHandler):
    pass


class SetResultSleepHandler(SleepHandler):
    def handle(self, *tasks: Task):
        super().handle(*tasks)
        for task in tasks:
            task.result = 'test'


@pytest.fixture
def log_file():
    f = tempfile.NamedTemporaryFile()
    yield f
    f.close()


@pytest.fixture
def aqueduct_logger(log_file):
    logger = logging.getLogger(LOGGER_NAME)
    logger.addHandler(logging.FileHandler(log_file.name))


@pytest.fixture
def sleep_handlers() -> Tuple[SleepHandler, ...]:
    return (
        SleepHandler1(0.001),
        SleepHandler2(0.003),
        SetResultSleepHandler(0.002),
    )


@pytest.fixture
def slow_sleep_handlers() -> Tuple[SleepHandler, ...]:
    return (
        SleepHandler1(0.1),
        SleepHandler2(0.3),
        SetResultSleepHandler(0.2),
    )

@pytest.fixture
def slow_priority_queue_handlers() -> Tuple[SleepHandler, ...]:
    return (
        SleepHandler1(0.1),
        SleepHandler2(0.1),
        SetResultSleepHandler(0.1),
    )


@asynccontextmanager
async def run_flow(flow: Flow):
    flow.start()
    yield flow
    await flow.stop(graceful=False)


@pytest.fixture
async def simple_flow(loop, sleep_handlers: Tuple[SleepHandler, ...]) -> Flow:
    async with run_flow(Flow(*sleep_handlers)) as flow:
        yield flow


@pytest.fixture
async def slow_simple_flow(loop, slow_sleep_handlers: Tuple[SleepHandler, ...]) -> Flow:
    async with run_flow(Flow(*slow_sleep_handlers)) as flow:
        yield flow

@pytest.fixture
async def slow_priority_queue_flow(loop, slow_priority_queue_handlers: Tuple[SleepHandler, ...]) -> Flow:
    async with run_flow(Flow(*slow_priority_queue_handlers, queue_priorities=2)) as flow:
        yield flow


@pytest.fixture
async def handle_condition_flow(loop) -> Flow:
    steps = (
        FlowStep(
            SleepHandler1(0.001),
            handle_condition=lambda task: task._type == 0
        ),
        FlowStep(
            SleepHandler2(0.001),
            handle_condition=lambda task: task._type == 1
        ),
        FlowStep(
            SetResultSleepHandler(0.001),
            handle_condition=lambda task: task._type in [0, 1]
        ),
    )
    async with run_flow(Flow(*steps)) as flow:
        yield flow



import asyncio
from dataclasses import asdict, dataclass
from multiprocessing import Queue
from typing import Dict, Iterable, Tuple

from . import IMetricsItems
from .base import MetricsItems, MetricsTypes
from .processes import ProcessesStats
from .task import TasksMetricsStorage


@dataclass
class TasksStats(IMetricsItems):
    cancel: int = 0
    timeout: int = 0
    complete: int = 0

    @property
    def items(self) -> Iterable[Tuple[str, int]]:
        return asdict(self).items()

    def extend(self, stats: 'TasksStats'):
        self.cancel += stats.cancel
        self.timeout += stats.timeout
        self.complete += stats.complete


class AqueductMetricsStorage(TasksMetricsStorage):
    def __init__(self):
        super().__init__()
        self.queue_sizes = MetricsItems()
        self.tasks_stats = TasksStats()
        self.memory_usage = MetricsItems()
        self.processes_stats = ProcessesStats()

    def extend(self, storage: TasksMetricsStorage):
        super().extend(storage)
        if isinstance(storage, AqueductMetricsStorage):
            self.queue_sizes.extend(storage.queue_sizes)
            self.tasks_stats.extend(storage.tasks_stats)
            self.processes_stats.extend(storage.processes_stats)

    def extend_memory_usage(self, metrics: MetricsItems):
        self.memory_usage.extend(metrics)


class Collector:
    def __init__(self, collectible_metrics: Iterable[MetricsTypes] = None,
                 qsize_collect_period: float = 1.):
        self._metrics = AqueductMetricsStorage()
        self._collectible_metrics = [MetricsTypes(t) for t in collectible_metrics] \
            if collectible_metrics is not None else list(MetricsTypes)

        self._qsize_collect_period: float = qsize_collect_period

    def is_collectible(self, metrics_type: MetricsTypes):
        return metrics_type in self._collectible_metrics

    async def collect_qsize(self, queues_info: Dict[Queue, str]):
        if not self.is_collectible(MetricsTypes.QUEUE_SIZE):
            return
        while True:
            queues_sizes = [(name, q.qsize()) for q, name in queues_info.items()]
            self._metrics.queue_sizes.add_items(queues_sizes)
            await asyncio.sleep(self._qsize_collect_period)

    def add_task_metrics(self, metrics: TasksMetricsStorage):
        self._metrics.extend(metrics)

    def add_tasks_stats(self, stats: TasksStats):
        self._metrics.tasks_stats.extend(stats)

    def add_memory_usage(self, metrics: MetricsItems):
        self._metrics.extend_memory_usage(metrics)

    def add_processes_stats(self, stats: ProcessesStats):
        self._metrics.processes_stats.extend(stats)

    def extract_metrics(self) -> AqueductMetricsStorage:
        metrics = self._metrics
        self._metrics = AqueductMetricsStorage()
        return metrics



"""Calculates the maximum rps for a service on a single physical node.

Calculations are performed when the service configuration is automatically changed.

To start, see README.md.
"""
import json
import pathlib
import subprocess
import sys
import tempfile
import time
from contextlib import contextmanager
from dataclasses import dataclass, field
from enum import Enum
from typing import List, Tuple

import logging

from .base import IMAGES, ReportRow, RpsCounterCfg
from .utils import run_cmd

PROJ_ROOT = pathlib.Path(__file__).parent.parent
REPORT_NAME = 'report_%s.csv'

RUN_SERVICE = f'docker-compose --env-file %s up -d --scale maas=%s maas'
STOP_SERVICE = 'docker-compose stop maas'
RM_SERVICE = 'docker-compose rm -f maas'
RUN_BENCH = 'docker-compose exec maas-bench python -m bench.rps_counter %s %s'
GET_SERVICE_IDS = 'docker ps --filter name=bench_service_maas -q'


class ScaleMethods(Enum):
    REPLICAS = 'replicas'
    PROC = 'proc'


@dataclass
class Cfg:
    path: str
    image_size: str
    image_name: str = field(init=False)
    replicas: int

    app_type: str
    nprocs: List[int]

    timeout: float
    load_time_secs: int
    success_rate: float
    min_rps: int
    scale_method: ScaleMethods

    def __post_init__(self):
        self.image_name = IMAGES[self.image_size]
        if self.app_type == 'pipeline':
            assert not self.nprocs
            assert self.scale_method == ScaleMethods.REPLICAS


def get_cfg(cfg_path: str) -> Cfg:
    with open(cfg_path) as f:
        cfg_info = json.loads(f.read())
    cfg_info['scale_method'] = ScaleMethods(cfg_info['scale_method'])
    return Cfg(cfg_path, **cfg_info)


@contextmanager
def running_service(app_type: str, replicas: int, nprocs: List[int]):
    with tempfile.NamedTemporaryFile('w', dir=PROJ_ROOT) as cfg_f:
        cfg_name = cfg_f.name.split('/')[-1]
        cfg_f.write(json.dumps({'app_type': app_type, 'nprocs': nprocs}))
        cfg_f.seek(0)

        with tempfile.NamedTemporaryFile('w', dir=PROJ_ROOT) as dc_env_f:
            dc_env_f.write(f'CFG_PATH={cfg_name}')
            dc_env_f.seek(0)

            run_cmd(RUN_SERVICE % (dc_env_f.name, replicas), shell=True)
            time.sleep(3)
            try:
                yield
            finally:
                run_cmd(STOP_SERVICE)
                run_cmd(RM_SERVICE)


def run_rps_counter(cfg: Cfg, replicas: int, nprocs: List[int], min_rps: int, report_name: str):
    container_ids = run_cmd(GET_SERVICE_IDS, stdout=subprocess.PIPE).stdout.read().decode().split()
    assert len(container_ids) == replicas, f'number of containers is {len(container_ids)} instead of {replicas}'
    rps_cfg = RpsCounterCfg(
        replicas=replicas,
        nprocs=nprocs,
        container_ids=container_ids,
        image_name=cfg.image_name,
        timeout=cfg.timeout,
        load_time_secs=cfg.load_time_secs,
        success_rate=cfg.success_rate,
        min_rps=min_rps or cfg.min_rps,
    )

    with tempfile.NamedTemporaryFile('w', dir=PROJ_ROOT) as f:
        f.write(json.dumps(rps_cfg.__dict__))
        f.seek(0)
        rps_cfg_name = f.name.split('/')[-1]
        run_cmd(RUN_BENCH % (rps_cfg_name, report_name))
    time.sleep(1)


def scaled(replicas: int, nprocs: List[int], scale_method: ScaleMethods) -> Tuple[int, List[int]]:
    # todo нужно дать возможность запускать сильно больше реплик (когда увеличивать кратно??)
    #  нужно отслеживать, что часть реплик не умерла к моменту теста (gpu oom)
    k = 1
    start_replicas, start_nprocs = replicas, nprocs
    while True:
        if scale_method == ScaleMethods.REPLICAS:
            replicas = start_replicas + k - 1
        else:
            nprocs = [n * k for n in start_nprocs]
        yield replicas, nprocs
        k += 1


def main(cfg_path: str):
    cfg = get_cfg(cfg_path)

    report_name = REPORT_NAME % cfg_path.split('/')[-1].split('.')[0]
    with open(report_name, 'w') as f:
        f.write(','.join(ReportRow.get_field_names()) + '\n')

    max_rps, failed_cnt = 0, 0

    for replicas, nprocs in scaled(cfg.replicas, cfg.nprocs, cfg.scale_method):
        with running_service(cfg.app_type, replicas, nprocs):
            run_rps_counter(cfg, replicas, nprocs, max_rps, report_name)

        max_rps_info = ReportRow.get_from_csv(
            report_name,
            lambda x: x.is_max_rps and x.replicas == replicas and x.nprocs == nprocs
        )
        if max_rps_info and max_rps_info[0].rps > max_rps:
            max_rps = max_rps_info[0].rps
        else:
            failed_cnt += 1
            if failed_cnt > 1:
                break

    logging.info(f'Absolutely max rps = {max_rps}')


if __name__ == '__main__':
    main(sys.argv[1])



"""
Copied from torch.multiprocessing.spawn v1.7.0
Added:
- inherits logger directly `logger = log.getChild('multiprocessing')`
- logs errors (e.g. for sending to Sentry)
"""

import multiprocessing
import multiprocessing.connection
import signal
import sys
import time
from typing import List, Optional

from .exceptions import AqueductError
from .logger import log

logger = log.getChild('multiprocessing')


class ProcessException(AqueductError):
    __slots__ = ['error_index', 'error_pid']

    def __init__(self, msg: str, error_index: int, pid: int):
        super().__init__(msg)
        self.error_index = error_index
        self.pid = pid


class ProcessRaisedException(ProcessException):
    """
    Exception is thrown when the process failed due to exception
    raised by the code.
    """
    def __init__(
        self,
        msg: str,
        error_index: int,
        error_pid: int,
    ):
        super().__init__(msg, error_index, error_pid)


class ProcessExitedException(ProcessException):
    """
    Exception is thrown when the process failed due to signal
    or exited with a specific code.
    """
    __slots__ = ['exit_code']

    def __init__(
            self, msg: str, error_index: int, error_pid: int,
            exit_code: int, signal_name: Optional[str] = None
    ):
        super().__init__(msg, error_index, error_pid)
        self.exit_code = exit_code
        self.signal_name = signal_name


def _wrap(fn, i, args, error_queue):
    try:
        fn(i, *args)
    except KeyboardInterrupt:
        pass  # SIGINT; Killed by parent, do nothing
    except Exception:
        # Propagate exception to parent process, keeping original traceback
        import traceback
        logger.exception(traceback.format_exc())
        error_queue.put(traceback.format_exc())
        sys.exit(1)


class ProcessContext:
    def __init__(self, processes: List[multiprocessing.process.BaseProcess],
                 error_queues: List[multiprocessing.SimpleQueue]):
        self.error_queues = error_queues
        self.processes = processes
        self.sentinels = {
            process.sentinel: index
            for index, process in enumerate(processes)
        }

    def __repr__(self):
        process = ','.join(map(str, self.processes))
        return f'<ProcessContext processes: {process}>'

    def pids(self):
        return [int(process.pid) for process in self.processes]

    def join(self, timeout=None):
        r"""
        Tries to join one or more processes in this process context.
        If one of them exited with a non-zero exit status, this function
        kills the remaining processes and raises an exception with the cause
        of the first process exiting.

        Returns ``True`` if all processes have been joined successfully,
        ``False`` if there are more processes that need to be joined.

        Arguments:
            timeout (float): Wait this long before giving up on waiting.
        """
        # Ensure this function can be called even when we're done.
        if len(self.sentinels) == 0:
            return True

        # Wait for any process to fail or all of them to succeed.
        ready = multiprocessing.connection.wait(
            self.sentinels.keys(),
            timeout=timeout,
        )

        error_index = None
        for sentinel in ready:
            index = self.sentinels.pop(sentinel)
            process = self.processes[index]
            process.join()
            if process.exitcode != 0:
                error_index = index
                break

        # Return if there was no error.
        if error_index is None:
            # Return whether or not all processes have been joined.
            return len(self.sentinels) == 0

        # Assume failure. Terminate processes that are still alive.
        for process in self.processes:
            if process.is_alive():
                process.terminate()
            process.join()

        # There won't be an error on the queue if the process crashed.
        failed_process = self.processes[error_index]
        if self.error_queues[error_index].empty():
            exitcode = self.processes[error_index].exitcode
            if exitcode < 0:
                name = signal.Signals(-exitcode).name
                raise ProcessExitedException(
                    'process %d terminated with signal %s' %
                    (error_index, name),
                    error_index=error_index,
                    error_pid=failed_process.pid,
                    exit_code=exitcode,
                    signal_name=name
                )
            else:
                raise ProcessExitedException(
                    'process %d terminated with exit code %d' %
                    (error_index, exitcode),
                    error_index=error_index,
                    error_pid=failed_process.pid,
                    exit_code=exitcode
                )

        original_trace = self.error_queues[error_index].get()
        msg = '\n\n-- Process %d terminated with the following error:\n' % error_index
        msg += original_trace
        raise ProcessRaisedException(msg, error_index, failed_process.pid)


def start_processes(fn, args=(), nprocs=1, join=True, daemon=False,
                    start_method='spawn', on_start_wait: float = 0):
    r"""Starts ``nprocs`` processes that run ``fn`` with ``args``.

    If one of the processes exits with a non-zero exit status, the
    remaining processes are killed and an exception is raised with the
    cause of termination. In the case an exception was caught in the
    child process, it is forwarded and its traceback is included in
    the exception raised in the parent process.

    Arguments:
        fn (function): Function is called as the entrypoint of the
            started process. This function must be defined at the top
            level of a module so it can be pickled and started. This
            is a requirement imposed by multiprocessing.

            The function is called as ``fn(i, *args)``, where ``i`` is
            the process index and ``args`` is the passed through tuple
            of arguments.

        args (tuple): Arguments passed to ``fn``.
        nprocs (int): Number of processes to start.
        join (bool): Perform a blocking join on all processes.
        daemon (bool): The started processes' daemon flag. If set to True,
                       daemonic processes will be created.
        start_method (string): Start method.
        on_start_wait (float): Timeout in sec to wait for all subprocesses to start

    Returns:
        None if ``join`` is ``True``,
        :class:`~ProcessContext` if ``join`` is ``False``

    """
    mp = multiprocessing.get_context(start_method)
    error_queues = []
    processes = []
    for i in range(nprocs):
        error_queue = mp.SimpleQueue()
        process = mp.Process(
            target=_wrap,
            args=(fn, i, args, error_queue),
            daemon=daemon,
        )
        process.start()
        error_queues.append(error_queue)
        processes.append(process)

    context = ProcessContext(processes, error_queues)
    if not join:
        # to prevent the parent process from exiting before the child processes have started
        if nprocs > 1 and start_method != 'fork':
            if on_start_wait != 0.0:
                time.sleep(on_start_wait)
            else:
                logger.warning('on_start_wait is not set, '
                               'may cause FileNotFoundError for heavy weight subprocess.')
        return context

    # Loop on join until it returns True or raises an exception.
    while not context.join():
        pass



import multiprocessing as mp
from dataclasses import dataclass

from typing import List

from .handler import HandleConditionType
from .task import BaseTask


@dataclass
class FlowStepQueue:
    queue: mp.Queue
    handle_condition: HandleConditionType


def select_next_queue(queues: List[List[FlowStepQueue]], task: BaseTask, start_index: int = 0) -> mp.Queue:
    for step_queue in queues[task.priority][start_index:]:
        if step_queue.handle_condition(task):
            return step_queue.queue
    return queues[task.priority][-1].queue



from multiprocessing import (
    Process,
    Queue,
)
from multiprocessing.shared_memory import SharedMemory

import numpy as np
import pytest

from aqueduct.shm.base import AtomicCounter
from aqueduct.shm.numpy import NPArraySharedData

from tests.unit.conftest import Task


class ArrayFieldTask(Task):
    def __init__(self, array: np.array, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.array = array


def update_sh_array(q_in: Queue, q_out: Queue):
    array_sh_data, value = q_in.get()
    sh_array = array_sh_data.get_data()
    sh_array[0] = value
    q_out.put((None, None))
    # дожидаемся изменения массива в главном процессе
    _, new_value = q_in.get()
    # теперь первый элемент массива равен new_value
    # сама проверка делается в родительском процессе, потому что исключение из дочернего не прокидывается
    # в родительский процесс автоматически
    q_out.put((sh_array[0] == new_value, f'array elem {sh_array[0]} is not equal to value {new_value}'))


def update_task_field(q_in: Queue, q_out: Queue):
    task, field_name, value = q_in.get()
    array = getattr(task, field_name)
    array[0] = value
    q_out.put(None)


@pytest.fixture
def task(array: np.ndarray):
    return ArrayFieldTask(array)


def handle_task(q_in: Queue, q_out: Queue):
    # была присоединена разделяемая память
    task = q_in.get()
    q_out.put(None)

    # объект, который ссылался на резделяемую память, должен быть удален gc
    task = q_in.get()
    q_out.put(None)

    q_in.get()


class TestNPArraySharedData:
    def test_access_to_data(self, array_sh_data: NPArraySharedData):
        sh_array = array_sh_data.get_data()

        q_in, q_out = Queue(), Queue()
        p = Process(target=update_sh_array, args=(q_out, q_in))
        p.start()

        value = sh_array[0] + 1
        assert sh_array[0] != value
        q_out.put((array_sh_data, value))
        # дожидаемся изменения массива в дочернем процессе
        _, _ = q_in.get()
        # теперь первый элемент массива равен value
        assert sh_array[0] == value
        value += 1
        sh_array[0] = value
        q_out.put((None, value))
        assert_expr, msg = q_in.get()
        assert assert_expr, msg

        p.join()


class TestSharedFieldsMixin:
    def test_pickle_task(self, task: ArrayFieldTask):
        field_name = 'array'

        state = task.__getstate__()
        assert field_name in state
        assert state['_shared_fields'] == {}

        # если поле расшарено, то оно не пиклится, вместо него пиклится мета информация
        task.share_value(field_name)
        state = task.__getstate__()
        assert field_name not in state
        assert field_name in state['_shared_fields']

    def test_update_task_field(self, task: ArrayFieldTask):
        task.share_value('array')

        q_in, q_out = Queue(), Queue()
        p = Process(target=update_task_field, args=(q_out, q_in))
        p.start()

        value = task.array[0] + 1
        q_out.put((task, 'array', value))
        # дожидаемся изменения массива в дочернем процессе
        _ = q_in.get()
        # теперь первый элемент массива равен value
        assert task.array[0] == value

        p.join()

    def test_auto_change_ref_counter(self, task: ArrayFieldTask):
        task.share_value('array')
        shf_info = task._shared_fields['array']
        # при присоединении памяти был увеличен счетчик ссылок
        assert shf_info.shm_wrapper.ref_count == 1

        q_in, q_out = Queue(), Queue()
        p = Process(target=handle_task, args=(q_out, q_in))
        p.start()

        q_out.put(task)
        q_in.get()
        # после присоединения области памяти дочерним процессом был увеличен счетчик ссылок
        assert shf_info.shm_wrapper.ref_count == 2

        q_out.put(None)
        q_in.get()
        # после работы gc в дочернем процессе счетчик ссылок был уменьшен
        assert shf_info.shm_wrapper.ref_count == 1

        q_out.put(None)
        p.join()

    def test_save_shared_value_directly(self, task: ArrayFieldTask):
        # НЕ рекомендуемый способ расшарить значение поля, но так тоже можно
        field_name = 'array'

        shared_array = NPArraySharedData.create_from_data(task.array)
        task.array = shared_array
        assert field_name in task._shared_fields
        assert task._shared_fields[field_name].shm_wrapper.ref_count == 1


class TestAtomicCounter:
    def test_size(self):
        assert AtomicCounter.size() == 4

    def test_init(self):
        view = memoryview(bytearray((1, 0, 0, 0)))
        AtomicCounter(view)

    def test_inc(self):
        view = memoryview(bytearray([0] * 16))
        c = AtomicCounter(view)

        assert c.inc() == 1
        assert c.inc() == 2
        assert c.dec() == 1
        assert c.dec() == 0

    @staticmethod
    def process_atomic(input):
        myshm = SharedMemory(name=input[0])
        myc = AtomicCounter(myshm.buf)

        if input[1] % 2 == 0:
            for _ in range(10000):
                myc.inc()
        if input[1] % 2 != 0:
            for _ in range(10000):
                myc.dec()

    def test_multiproc_atomic(self):
        """Modify counter in multiple processes.

        There is equal amount of increments/decrements - so result should be 0
        """
        from multiprocessing import Pool
        shm = SharedMemory(create=True, size=1024 * 1024 + AtomicCounter.size())
        c = AtomicCounter(shm.buf)

        pool = Pool(10)
        pool.map(self.process_atomic, [(shm.name, i) for i in range(10)])

        assert c.get() == 0



import asyncio
import multiprocessing as mp
import operator
import os
import psutil
import queue
import signal
import sys
from enum import Enum
from functools import cached_property
from functools import reduce
from itertools import chain
from multiprocessing import Barrier
from threading import BrokenBarrierError
from time import monotonic
from typing import Dict, List, Literal, Optional, Union

from concurrent.futures import ThreadPoolExecutor
from multiprocessing.resource_tracker import _resource_tracker  # noqa

from .exceptions import FlowError, MPStartMethodValueError, NotRunningError
from .handler import BaseTaskHandler, HandleConditionType
from .logger import log
from .metrics import MAIN_PROCESS, MetricsTypes
from .metrics.collect import Collector, TasksStats
from .metrics.export import Exporter
from .metrics.manager import get_metrics_manager
from .metrics.processes import ProcessesStats
from .metrics.queue import TaskMetricsQueue
from .metrics.timer import timeit
from .multiprocessing import (
    ProcessContext,
    ProcessExitedException,
    ProcessRaisedException,
    start_processes,
)
from .queues import FlowStepQueue, select_next_queue
from .task import BaseTask, DEFAULT_PRIORITY, StopTask
from .worker import Worker
from .metrics.base import MetricsItems

# just for using common ResourceTracker in main and child processes and avoiding
# unnecessary shared memory resource_tracker "No such file or directory" warnings
_resource_tracker.ensure_running()


def _check_env():
    if not sys.version_info >= (3, 8):
        raise RuntimeError('Requires python 3.8 or higher to use multiprocessing.shared_memory')


class FlowStep:

    def __init__(
            self,
            handler: BaseTaskHandler,
            handle_condition: HandleConditionType = operator.truth,
            nprocs: int = 1,
            batch_size: int = 1,
            batch_timeout: float = 0,
            on_start_wait: float = 0,
    ):
        _check_env()
        self.handler = handler
        self.handle_condition = handle_condition
        self.nprocs = nprocs
        self.batch_size = batch_size
        self.batch_timeout = batch_timeout
        self.on_start_wait = on_start_wait


class FlowState(Enum):
    RUNNING = 'running'
    STARTING = 'starting'
    STOPPING = 'stopping'
    STOPPED = 'stopped'


class Flow:

    def __init__(
            self,
            *steps: Union[FlowStep, BaseTaskHandler],
            metrics_enabled: bool = True,
            metrics_collector: Collector = None,
            metrics_exporter: Exporter = None,
            queue_size: Optional[int] = None,
            queue_priorities: int = 1,
            mp_start_method: Literal['fork', 'spawn', 'forkserver'] = 'fork',
    ):
        _check_env()

        self._steps: List[FlowStep] = [step if isinstance(step, FlowStep)
                                       else FlowStep(step) for step in steps]
        self._contexts: Dict[BaseTaskHandler, ProcessContext] = {}
        self._queue_priorities = queue_priorities
        self._queues: List[List[FlowStepQueue]] = []
        self._task_futures: Dict[str, asyncio.Future] = {}
        self._queue_size: Optional[int] = queue_size

        if mp_start_method != "fork" and mp_start_method != mp.get_start_method():
            log.error(f'MP start method {mp_start_method!r} is set for Flow, it should also be set'
                      f' in the if __name__ == "__main__" clause of the main module')
            raise MPStartMethodValueError(f'Multiprocessing start method mismatch: '
                                          f'got {mp.get_start_method()!r} for main process '
                                          f'and {mp_start_method!r} for Flow')
        self._mp_start_method = mp_start_method

        if not metrics_enabled:
            log.warn('Metrics collecting is disabled')
            metrics_collector = Collector(collectible_metrics=[])
        self._metrics_manager = get_metrics_manager(metrics_collector, metrics_exporter)

        self._state: FlowState = FlowState.STOPPED
        self._tasks: List[asyncio.Future] = []

    @property
    def state(self):
        return self._state

    @property
    def is_running(self) -> bool:
        return self._state == FlowState.RUNNING

    def start(self, timeout: Optional[int] = None):
        """
        Starts Flow and waits for all subprocesses to initialize.

        raising FlowError if 'timeout' is set and some handler was not able to initialize in time.
        """
        log.info('Flow is starting')
        self._state = FlowState.STARTING
        self._run_steps(timeout)
        self._run_tasks()
        self._state = FlowState.RUNNING
        log.info('Flow was started')

    async def process(self, task: BaseTask, timeout_sec: float = 5.) -> bool:
        """
        Starts processing task through pipeline and returns True if all was ok.
        The result is saved in the source task
        """

        if not self.is_running:
            raise NotRunningError

        with timeit() as timer:
            future = asyncio.Future()
            self._task_futures[task.task_id] = future

            task.priority = min(task.priority, self._queue_priorities - 1)
            task.set_timeout(timeout_sec)
            task.metrics.start_transfer_timer(MAIN_PROCESS)

            start_time = monotonic()
            to_queue = select_next_queue(
                queues=self._queues,
                task=task,
            )
            while self.state != FlowState.STOPPED and (monotonic() - start_time) < timeout_sec:
                try:
                    to_queue.put(task, block=False)
                except queue.Full:
                    await asyncio.sleep(0.001)
                else:
                    break

            elapsed_time = monotonic() - start_time

            tasks_stats = TasksStats()
            try:
                finished_task: BaseTask = await asyncio.wait_for(
                    future,
                    timeout=(timeout_sec - elapsed_time),
                )
            # todo is it correct to hide a specific error behind a general FlowError?
            except asyncio.TimeoutError:
                tasks_stats.timeout += 1
                raise FlowError('Task timeout error')
            except asyncio.CancelledError:
                tasks_stats.cancel += 1
                if self.state in (FlowState.STOPPING, FlowState.STOPPED):
                    raise FlowError('Task was cancelled')
                else:
                    # process was cancelled by external actor, so reraise
                    raise

            else:
                tasks_stats.complete += 1
            finally:
                self._metrics_manager.collector.add_tasks_stats(tasks_stats)
                del self._task_futures[task.task_id]

        finished_task.metrics.handle_times.add('total', timer.seconds)

        if self.need_collect_task_timers:
            self._metrics_manager.collector.add_task_metrics(finished_task.metrics)

        task.update(finished_task)

        return True

    async def stop(self, graceful: bool = True):
        if not self.is_running:
            log.info('Flow is not running')
            return

        self._state = FlowState.STOPPING
        log.info('Flow is stopping')

        if graceful:
            self._queues[DEFAULT_PRIORITY][0].queue.put(StopTask())
            await asyncio.sleep(3)

        self._metrics_manager.stop()
        for task in self._tasks:
            task.cancel()
        for task in self._task_futures.values():
            task.cancel()

        self._join_context(self._processes_context)

        self._state = FlowState.STOPPED
        log.info('Flow was stopped')

    @cached_property
    def need_collect_task_timers(self) -> bool:
        return self._metrics_manager.collector.is_collectible(MetricsTypes.TASK_TIMERS)

    def _calc_queue_size(self, step: FlowStep):
        """ If queue size not specified manually, get queue size based on batch size for handler.
        We need at least batch_size places in queue and then some additional space
        """
        if self._queue_size is not None:
            return self._queue_size

        # queue should be able to store at least 20 task, that's seems reasonable
        return max(step.batch_size*3, 20)
    
    async def _check_memory_usage(self, sleep_sec: float = 1.):
        handler_processes_dict = {}
        for step_number, handler in enumerate(self._contexts):
            flow_step_name = handler.get_step_name(step_number)
            pids = self._contexts[handler].pids()
            processes = []
            for pid in pids:
                process = psutil.Process(pid)
                processes.append(process)
            handler_processes_dict[flow_step_name] = processes

        while self.state != FlowState.STOPPED:
            metrics = MetricsItems()
            all_memory_usage = 0
            for flow_step_name, processes in handler_processes_dict.items():
                nprocs_memory_sum = 0
                for process in processes:
                    memory = process.memory_info().rss
                    nprocs_memory_sum  += memory
                    metrics.add(flow_step_name, memory)
                all_memory_usage += nprocs_memory_sum
                if len(processes) != 1:
                    metrics.add(f'{flow_step_name}_nprocs_sum', nprocs_memory_sum)
            metrics.add('all_memory_usage', all_memory_usage)
            self._metrics_manager.collector.add_memory_usage(metrics)
            await asyncio.sleep(sleep_sec)

    def _run_steps(self, timeout: Optional[int]):
        if len(self._steps) == 0:
            log.info('Flow has zero steps -> do nothing')
            return

        # count how many processes we will create, to setup a barrier
        total_procs = reduce(lambda a, b: a + b.nprocs, self._steps, 0)
        # also count main process
        total_procs += 1
        start_barrier = Barrier(total_procs)

        for _ in range(self._queue_priorities):
            queues: List[FlowStepQueue] = []
            for step in self._steps:
                queue_size = self._calc_queue_size(step)
                queues.append(
                    FlowStepQueue(
                        queue=TaskMetricsQueue(queue_size),
                        handle_condition=step.handle_condition,
                    )
                )

            # Add out queue
            queue_size = self._calc_queue_size(self._steps[-1])
            queues.append(
                FlowStepQueue(
                    queue=TaskMetricsQueue(queue_size),
                    handle_condition=operator.truth,
                )
            )
            self._queues.append(queues)

        for step_number, step in enumerate(self._steps, 1):
            worker_curr = Worker(
                queues=self._queues,
                task_handler=step.handler,
                batch_size=step.batch_size,
                batch_timeout=step.batch_timeout,
                batch_lock=mp.RLock() if step.nprocs > 1 and step.batch_size > 1 else None,
                read_lock=mp.RLock(),
                step_number=step_number,
            )
            self._contexts[step.handler] = start_processes(
                worker_curr.loop,
                nprocs=step.nprocs, join=False, daemon=True, start_method=self._mp_start_method,
                args=(start_barrier,),
                on_start_wait=step.on_start_wait,
            )
            log.info(f'Created step {step.handler}, '
                     f'queue_in: {self._queues[DEFAULT_PRIORITY][step_number - 1].queue}, '
                     f'queue_out: {self._queues[DEFAULT_PRIORITY][step_number].queue}')

        # fix to avoid deadlock on program exit
        for step_queue in chain.from_iterable(self._queues):
            step_queue.queue.cancel_join_thread()

        try:
            log.info(f'Waiting for all workers to startup for {timeout} seconds...')
            start_barrier.wait(timeout)
        except BrokenBarrierError:
            raise TimeoutError('Starting timeout expired')

    def _run_tasks(self):
        self._tasks.append(asyncio.ensure_future(self._fetch_processed()))
        self._tasks.append(asyncio.ensure_future(self._check_is_alive()))

        self._metrics_manager.start(queues_info=self._get_queues_info())
        self._tasks.append(asyncio.ensure_future(self._check_memory_usage()))

    def _get_queues_info(self) -> Dict[mp.Queue, str]:
        """Returns queues between Step handlers and its names.

        The queue name consists of two handler names that are connected by this queue.
        """

        def step_names(handlers):
            from_step = MAIN_PROCESS
            for step_number, handler in enumerate(handlers, 1):
                to_step = handler.get_step_name(step_number)
                yield from_step, to_step
                from_step = to_step
            yield from_step, MAIN_PROCESS
        result = {}
        for priority in range(self._queue_priorities):
            result.update(
                {
                    queue_.queue: f'p_{priority}_from_{from_}_to_{to}' if priority > 0 else f'from_{from_}_to_{to}'
                    for queue_, (from_, to) in zip(self._queues[priority], step_names(self._contexts))
                }
            )
        return result

    @staticmethod
    def _fetch_from_queue(out_queue: mp.Queue) -> Union[BaseTask, None]:
        try:
            task = out_queue.get(timeout=1.)
            return task
        except queue.Empty:
            return None

    def _read_from_queue(self, loop: asyncio.AbstractEventLoop, q: mp.Queue) -> None:
        while self.state != FlowState.STOPPED:
            task = self._fetch_from_queue(q)

            if task is None:
                continue

            fut = self._task_futures.get(task.task_id)
            if fut and not fut.cancelled() and not fut.done():
                task.metrics.stop_transfer_timer(MAIN_PROCESS, task.priority)
                task_size = getattr(q, 'task_size', None)
                if task_size:
                    task.metrics.save_task_size(task_size, MAIN_PROCESS, task.priority)

                loop.call_soon_threadsafe(fut.set_result, task)

    async def _fetch_processed(self):
        """Fetching messages from output queue.

        To handle messages from another process and not block asyncio loop, we run queue.get()
        in a separate thread

        """
        loop = asyncio.get_event_loop()
        with ThreadPoolExecutor(max_workers=self._queue_priorities) as queue_fetch_executor:
            results_queues = [
                queues_with_same_priority[-1].queue for queues_with_same_priority in self._queues
            ]
            tasks = [
                loop.run_in_executor(queue_fetch_executor, self._read_from_queue, loop, q)
                for q in results_queues
            ]
            done, pending = await asyncio.wait(tasks, return_when=asyncio.ALL_COMPLETED)
            for t in done:
                try:
                    t.result()
                except Exception:
                    log.exception('error in queue-consumer')

            for t in pending:
                t.cancel()

    async def _check_is_alive(self, sleep_sec: float = 1.):
        """Checks that all child processes are alive.

        If at least one process is not alive, it stops Flow.
        """
        while self.state != FlowState.STOPPED:
            processes_stats = ProcessesStats()
            for handler, context in self._contexts.items():
                for proc in context.processes:
                    if not proc.is_alive():
                        if self.is_running:
                            handler_name = handler.__class__.__name__
                            log.error('The process %s for %s handler is dead',
                                      proc.pid, handler_name)
                            processes_stats.add_dead_process()
                            self._metrics_manager.collector.add_processes_stats(processes_stats)
                            await self.stop(graceful=False)
                    else:
                        processes_stats.add_running_process()
            self._metrics_manager.collector.add_processes_stats(processes_stats)
            await asyncio.sleep(sleep_sec)

    @staticmethod
    def _join_context(context: ProcessContext, timeout_sec: float = 0.01):
        try:
            context.join(timeout_sec)
        except (ProcessExitedException, ProcessRaisedException):
            pass

        for p in context.processes:
            if p.is_alive():
                os.kill(p.pid, signal.SIGKILL)

    @property
    def _processes_context(self) -> ProcessContext:
        processes, error_queues = [], []
        for context in self._contexts.values():
            processes.extend(context.processes)
            error_queues.extend(context.error_queues)
        return ProcessContext(processes=processes, error_queues=error_queues)



import asyncio
import pathlib
import sys
import tempfile
import time
from collections import Counter
from dataclasses import dataclass, field
from http import HTTPStatus
import logging
from typing import AsyncContextManager, Awaitable, Callable, List, Optional, Tuple

import aiohttp
import docker
import pandas as pd

from .base import ReportRow, RpsCounterCfg, get_rps_counter_cfg
from pulemet import Pulemet
from .utils import MaxNumberFinder, run_cmd

PROJ_ROOT = pathlib.Path(__file__).parent.parent

IMAGES_DIR = pathlib.Path(__file__).parent / 'data'
GPU_QUERY = 'nvidia-smi --query-gpu=%s -l 1 --format=csv'
MAAS_URL = 'http://maas:8080/classify'


def restart_service(ids: List[str], sleep_secs: int = None):
    client = docker.DockerClient(base_url='tcp://172.17.0.1:2376')
    for c in client.containers.list():
        if c.id in ids:
            c.restart()
    if sleep_secs:
        time.sleep(sleep_secs)


@dataclass
class RequestData:
    url: str
    timeout: float
    method: str = 'post'
    json: dict = field(default_factory=dict)
    data: bytes = None


class ReqStats(Counter):
    @property
    def success_rate(self) -> float:
        return self.get(HTTPStatus.OK, 0) / sum(self.values())

    @property
    def error_rate(self) -> float:
        return 1 - self.success_rate


@dataclass
class GPUStatsField:
    name: str
    type_: type
    units: str


GPU_UTIL = GPUStatsField('utilization.gpu', int, '%')
USED_MEM = GPUStatsField('memory.used', int, 'MiB')
MEM_UTIL = GPUStatsField('utilization.memory', int, '%')
POWER = GPUStatsField('power.draw', float, 'W')
GPU_STATS_FIELDS = (GPU_UTIL, USED_MEM, MEM_UTIL, POWER)


class NvidiaSmiStats:
    def __init__(self):
        self._field_names = [f.name for f in GPU_STATS_FIELDS]
        self._f = None
        self._proc = None
        self._df = None

    def _get_gpu_stats(self) -> pd.DataFrame:
        stats = pd.read_csv(self._f.name, names=self._field_names, header=0)
        for col, field in zip(stats, GPU_STATS_FIELDS):
            stats[col] = stats[col].apply(lambda x: field.type_(x.split()[0]))
        return stats

    def __enter__(self):
        self._f = tempfile.NamedTemporaryFile('w')
        self._proc = run_cmd(GPU_QUERY % ','.join(self._field_names), stdout=self._f, wait=False)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._proc.terminate()
        self._df = self._get_gpu_stats()
        self._f.close()

    @property
    def df(self) -> pd.DataFrame:
        return self._df


async def get_requests_stats(
        responses,
        check_ok: Callable[[aiohttp.ClientResponse], Awaitable[bool]] = None,
) -> ReqStats:
    statuses = []
    for resp in responses:
        if isinstance(resp, AsyncContextManager):
            async with resp as r:
                statuses.append(r.status)
                if check_ok:
                    if r.status == HTTPStatus.OK:
                        await check_ok(r)
        elif isinstance(resp, asyncio.TimeoutError):
            statuses.append(499)
        elif isinstance(resp, aiohttp.ClientConnectionError):
            statuses.append(HTTPStatus.SERVICE_UNAVAILABLE.value)
        else:
            raise Exception(f'Unknown response: {type(resp)}, {resp}')
    return ReqStats(statuses)


async def check_resp_body(r: aiohttp.ClientResponse):
    body = await r.json()
    assert isinstance(body, dict), f'Response body is not dict: {body}'
    assert isinstance(body.get('result'), list), f'Bad body: {body}'


async def send_requests(req: RequestData, rps: int, time_interval_secs: int) -> Tuple:
    async with aiohttp.ClientSession() as session:
        coros = [
            session.request(req.method, req.url, json=req.json, timeout=req.timeout)
            for _ in range(rps * time_interval_secs)
        ]
        pulemet = Pulemet(rps=rps, pool_size=10**6)
        pulemet_coros = pulemet.process(coros)

        return await asyncio.gather(*pulemet_coros, return_exceptions=True)


async def warmup_service(cfg: RpsCounterCfg):
    logging.info('start to warmup')
    for i in range(1, 4):
        req_stats, _ = await load_service(cfg.image_name, cfg.min_rps, cfg.load_time_secs, cfg.timeout)
        if req_stats.error_rate > 0:
            time.sleep(2**i)
            continue
        logging.info('warmup successfully finished')
        break
    else:
        raise Exception('warmup failed')


async def load_service(
        image_name: str,
        rps: int,
        time_interval_secs: int,
        timeout: float,
) -> Tuple[ReqStats, NvidiaSmiStats]:
    """Returns statistics of requests sent to the server."""
    logging.info(f'\nstart to load with rps = {rps} during {time_interval_secs} secs')
    r = RequestData(url=MAAS_URL, timeout=timeout, json={'image_name': image_name})
    with NvidiaSmiStats() as gpu_stats:
        responses = await send_requests(r, rps=rps, time_interval_secs=time_interval_secs)
    req_stats = await get_requests_stats(responses, check_ok=check_resp_body)
    logging.info(f'requests stats is {req_stats}, success rate is {req_stats.success_rate * 100:.1f}%')
    logging.info(
        f'max value of GPU_UTIL is {gpu_stats.df[GPU_UTIL.name].max()}{GPU_UTIL.units}, '
        f'mean value of GPU UTIL is {gpu_stats.df[GPU_UTIL.name].mean():.1f}{GPU_UTIL.units}'
    )
    return req_stats, gpu_stats


def get_report_row(rps: int, success_rate: float, cfg: RpsCounterCfg, gpu_stats: NvidiaSmiStats) -> ReportRow:
    return ReportRow(
        rps=rps,
        success_rate=success_rate,
        replicas=cfg.replicas,
        nprocs=cfg.nprocs,
        image_name=cfg.image_name,
        load_time_secs=cfg.load_time_secs,
        max_GPU_UTIL=gpu_stats.df[GPU_UTIL.name].max(),
        mean_GPU_UTIL=gpu_stats.df[GPU_UTIL.name].mean(),
    )


async def calculate_rps(cfg: RpsCounterCfg, report_path: str) -> Optional[int]:
    logging.info({k: v for k, v in cfg.__dict__.items() if k not in ('container_ids', 'image_name')})

    await warmup_service(cfg)

    report = pd.DataFrame(columns=ReportRow.get_field_names())
    with MaxNumberFinder(cfg.min_rps) as rps_counter:
        while rps_counter.number is not None:  # just for PyCharm, really it means `while True`
            req_stats, gpu_stats = await load_service(
                cfg.image_name, rps_counter.number, cfg.load_time_secs, cfg.timeout)
            report_row = get_report_row(rps_counter.number, req_stats.success_rate, cfg, gpu_stats)
            if req_stats.success_rate < cfg.success_rate:
                rps_counter.next(less=True)
                restart_service(cfg.container_ids, cfg.load_time_secs)
                await warmup_service(cfg)
            else:
                rps_counter.next()
            report = report.append(report_row.__dict__, ignore_index=True)
    report = report.append(report_row.__dict__, ignore_index=True)

    logging.info(f'max rps is {rps_counter.max_number}')
    report.loc[report['rps'] == rps_counter.max_number, 'is_max_rps'] = True
    report.to_csv(report_path, mode='a', header=False, index=False, float_format='%.3f')

    return rps_counter.max_number


if __name__ == '__main__':
    asyncio.run(calculate_rps(get_rps_counter_cfg(sys.argv[1]), sys.argv[2]))



"""Provides tools for managing data in shared memory.

There is a feature in shared_memory - even after close() in another process
where memory was linked, unlink() is called when exiting the process.
As a result, this piece of memory cannot be used anywhere else, because it was returned to the OS.
The reason is in the resource_tracker, which is separate for each non-child process.

You can bypass this feature by explicitly excluding a piece of memory from the tracker's observation:
    from multiprocessing.resource_tracker import unregister
    unregister(f'/{shm.name}', 'shared_memory')

Also, this problem does not exist if the process using memory is a child (because there is the same resource tracker)
https://bugs.python.org/issue39959
"""

from abc import (
    ABC,
    abstractmethod,
)

try:
    from multiprocessing import shared_memory
except ImportError:
    raise ImportError('shared_memory module is available since python3.8')
from typing import Any

from ._atomic import ffi
from ._atomic.lib import load_uint32, store_uint32, add_and_fetch_uint32, sub_and_fetch_uint32


class AtomicCounter:
    def __init__(self, view: memoryview):
        self._ptr = ffi.cast('uint32_t*', ffi.from_buffer(view[:self.size()]))

    def get(self):
        return load_uint32(self._ptr)

    def set(self, n):
        return store_uint32(self._ptr, n)

    def inc(self):
        return add_and_fetch_uint32(self._ptr, 1)

    def dec(self):
        return sub_and_fetch_uint32(self._ptr, 1)

    @staticmethod
    def size():
        return ffi.sizeof('uint32_t')


class SharedMemoryWrapper:
    """Manages the shared memory lifecycle.

    Allocates shared memory, manages reference counts and returns shared memory to OS.
    """

    def __init__(self, size: int, shm: shared_memory.SharedMemory = None, shm_name: str = None):
        self._shm: shared_memory.SharedMemory = None  # noqa

        if shm is None and shm_name is None:
            rc_size = AtomicCounter.size()
            shm = shared_memory.SharedMemory(create=True, size=size + rc_size)
            self._rc = AtomicCounter(shm.buf)
            self._rc.set(1)

        self._attach_shm(shm, shm_name)

        # it's necessary for deserialization
        self._shm_name = self._shm.name

    @property
    def buf(self) -> memoryview:
        return self._shm.buf[AtomicCounter.size():]

    @property
    def ref_count(self) -> int:
        return self._rc.get()

    def _attach_shm(self, shm: shared_memory.SharedMemory = None, shm_name: str = None):
        if shm is None:
            shm = shared_memory.SharedMemory(shm_name)
        self._shm = shm
        self._shm_name = shm.name

    def __getstate__(self) -> dict:
        self._rc.inc()
        state = {k: v for k, v in self.__dict__.items() if k not in ('_shm', '_rc')}
        return state

    def __setstate__(self, state: dict):
        self.__dict__.update(state)
        self._attach_shm(shm_name=self._shm_name)
        self._rc = AtomicCounter(self._shm.buf)

    def __del__(self):
        curr_rc = self._rc.dec()
        if curr_rc == 0:
            self._shm.unlink()


class SharedData(ABC):
    """Manages data in shared memory."""

    def __init__(self, shm_wrapper: SharedMemoryWrapper):
        self.shm_wrapper = shm_wrapper

    @abstractmethod
    def get_data(self) -> Any:
        """Returns a ref to the data in shared memory."""

    @classmethod
    @abstractmethod
    def create_from_data(cls, data: Any) -> 'SharedData':
        """Copies the data to shared memory and returns an instance of the class."""



import csv
import json
from dataclasses import dataclass
from typing import Callable, List


IMAGES = {
    'xs': 'apple.jpg',
    's': 'apples.jpg',  # small
    'm': 'cat.jpg',
    'l': 'umbrella.jpg',
}


@dataclass
class RpsCounterCfg:
    replicas: int
    nprocs: List[int]
    container_ids: List[str]
    image_name: str
    timeout: float
    load_time_secs: int
    success_rate: float
    min_rps: int


def get_rps_counter_cfg(cfg_path: str) -> RpsCounterCfg:
    with open(cfg_path) as f:
        cfg = json.loads(f.read())
    return RpsCounterCfg(**cfg)


@dataclass
class ReportRow:
    rps: int
    success_rate: float
    replicas: int
    nprocs: List[int]
    max_GPU_UTIL: float
    mean_GPU_UTIL: float
    image_name: str
    load_time_secs: int
    is_max_rps: bool = False

    @classmethod
    def get_field_names(cls):
        return list(cls.__dataclass_fields__.keys())

    @classmethod
    def get_from_csv(
            cls,
            csv_name: str,
            condition: Callable[['ReportRow'], bool] = lambda x: True,
    ) -> List['ReportRow']:
        lst = []
        with open(csv_name) as f:
            reader = csv.DictReader(f)
            for row in reader:
                row['is_max_rps'] = row['is_max_rps'].lower()
                image_name = row.pop('image_name')
                report_row = ReportRow(**{k: json.loads(v) for k, v in row.items()}, image_name=image_name)
                if condition(report_row):
                    lst.append(report_row)
        return lst



class AqueductError(Exception):
    """Base class for all Aqueduct errors."""


class FlowError(AqueductError):
    """Flow can raise this if something was wrong."""


class MPStartMethodValueError(FlowError, ValueError):
    """Flow will raise this if flow's method to start subprocess differs from the main method"""


class NotRunningError(FlowError):
    """Flow can raise this if it's not already running."""


class BadReferenceCount(AqueductError, ValueError):
    pass



import time
import pytest

from aqueduct.metrics.task import TaskMetrics
from aqueduct.metrics.timer import timeit


@pytest.fixture
def task_metrics():
    return TaskMetrics()


class TestTaskMetrics:
    def test_timeit(self, task_metrics):
        with timeit() as timer:
            time.sleep(0.001)
        task_metrics.handle_times.add('test_metric', timer.seconds)

        assert task_metrics.handle_times.items[0][1] >= 0.001



from typing import (
    Any,
    Dict,
    Protocol
)
import warnings

from .base import SharedData, SharedMemoryWrapper
from .bytes import BytesSharedData



class HasReadany(Protocol):
    """We need only single method from aiohttp.streams.StreamReader"""
    async def readany(self) -> bytes:
        pass


class SharedFieldsMixin:
    """Allows you to replace the value of a class instance field with its copy in shared memory."""

    def __init__(self, *args, **kwargs):
        self._shared_fields: Dict[str, SharedData] = {}
        super().__init__(*args, **kwargs)
    
    async def share_value_with_data(self, field_name: str, content: HasReadany, size: int):
        """Creates a shared memory in field_name attribute of type bytes of size
        and loads data there from a source that has async method readany
        """

        shm_wrapper = SharedMemoryWrapper(size)
        shared_value = BytesSharedData(shm_wrapper, size, bytes)

        offset = 0
        while True:
            block = await content.readany()
            if not block:
                break
            shm_wrapper.buf[offset: offset+len(block)] = block
            offset += len(block)
        
        self._set_shared_value(field_name, shared_value)

    def share_value(self, field_name: str):
        """Replaces the value of the field with its copy in shared memory.
           All subsequent changes to the value will be made in shared memory, i.e. these changes
           they will be reflected in all objects that refer to the same memory area.
           If the value is replaced, this method must be called again so that the value gets into
           shared memory, and a new area of shared memory will be allocated.
        """
        if field_name not in self.__dict__:
            raise ValueError('Unknown field name')
        if field_name in self._shared_fields:
            raise ValueError(f'The {field_name} field value is already shared')

        value = getattr(self, field_name)
        shared_value = self._get_shared_value(value)
        self._set_shared_value(field_name, shared_value)

    @staticmethod
    def _get_shared_value(value: Any) -> SharedData:
        try:
            import numpy as np
        except ImportError:
            if isinstance(value, (bytes, bytearray)):
                return BytesSharedData.create_from_data(value)
        else:
            from .numpy import NPArraySharedData
            if isinstance(value, np.ndarray):
                return NPArraySharedData.create_from_data(value)
            if isinstance(value, (bytes, bytearray)):
                return BytesSharedData.create_from_data(value)

        raise ValueError(f'Type {type(value)} cannot be shared')

    def _set_shared_value(self, field_name: str, shared_value: SharedData):
        self.__dict__[field_name] = shared_value.get_data()
        self._shared_fields[field_name] = shared_value

    def __setattr__(self, key, value):
        if isinstance(value, SharedData):
            self._set_shared_value(key, value)
        else:
            super().__setattr__(key, value)
            if key in self._shared_fields:
                self._shared_fields.pop(key)
                warnings.warn(f'The {key} field value is not shared anymore')

    def __getstate__(self) -> dict:
        state = {k: v for k, v in self.__dict__.items() if k not in self._shared_fields}
        return state

    def __setstate__(self, state: dict):
        self.__dict__.update(state)
        for f_name, shared_value in self._shared_fields.items():
            self.__dict__[f_name] = shared_value.get_data()



from .base import IExtendable, MetricsItems
from .timer import TransferTimer


class TasksMetricsStorage(IExtendable):
    def __init__(self):
        self.transfer_times = MetricsItems()
        self.task_sizes = MetricsItems()
        self.handle_times = MetricsItems()
        self.batch_times = MetricsItems()
        self.batch_sizes = MetricsItems()

    def extend(self, storage: 'TasksMetricsStorage'):
        self.transfer_times.extend(storage.transfer_times)
        self.task_sizes.extend(storage.task_sizes)
        self.handle_times.extend(storage.handle_times)
        self.batch_times.extend(storage.batch_times)
        self.batch_sizes.extend(storage.batch_sizes)


class TaskMetrics(TasksMetricsStorage):
    """Task's "backpack" with metrics.

    It is used to store metrics related to a task passing through the Flow child processes. This mechanic
    works as long as all tasks reach the output queue in the main process.
    """
    def __init__(self):
        super().__init__()
        self._transfer_timer: TransferTimer = None  # noqa

    def start_transfer_timer(self, transfer_from: str):
        self._transfer_timer = TransferTimer(transfer_from)
        self._transfer_timer.start()

    def stop_transfer_timer(self, transfer_to: str, priority: int = 0):
        self._transfer_timer.stop()
        from_ = self._transfer_timer.transfer_from
        name = (
            f'p_{priority}_from_{from_}_to_{transfer_to}'
            if priority > 0 else f'from_{from_}_to_{transfer_to}'
        )
        self.transfer_times.add(name, self._transfer_timer.seconds)

    def save_task_size(self, task_size: int, transfer_to: str, priority: int = 0):
        from_ = self._transfer_timer.transfer_from
        name = (
            f'p_{priority}_from_{from_}_to_{transfer_to}'
            if priority > 0 else f'from_{from_}_to_{transfer_to}'
        )
        self.task_sizes.add(name, task_size)


import time
from typing import Optional


class Timer:
    precision = 5

    def __init__(self):
        self._start = None
        self._stop = None

    @staticmethod
    def _time() -> float:
        return time.monotonic()

    def start(self):
        self._start = self._time()

    def stop(self):
        self._stop = self._time()

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.stop()

    @property
    def seconds(self) -> Optional[float]:
        if self._start is None or self._stop is None:
            return
        return round(self._stop - self._start, self.precision)


class TransferTimer(Timer):
    def __init__(self, transfer_from: str):
        self.transfer_from = transfer_from
        self.transfer_to: str = None  # noqa
        super().__init__()


timeit = Timer



from typing import Tuple

import numpy as np

from .base import SharedData, SharedMemoryWrapper


class NPArraySharedData(SharedData):
    def __init__(self, shm_wrapper: SharedMemoryWrapper, shape: Tuple[int, ...], dtype: np.dtype):
        self._shape = shape
        self._dtype = dtype
        super().__init__(shm_wrapper)

    def get_data(self) -> np.ndarray:
        if self.shm_wrapper.buf is None:
            raise ValueError('No shared memory buffer')
        return np.ndarray(self._shape, dtype=self._dtype, buffer=self.shm_wrapper.buf)

    @classmethod
    def create_from_data(cls, data: np.ndarray) -> 'NPArraySharedData':
        shm_wrapper = SharedMemoryWrapper(data.nbytes)
        shared_data = cls(shm_wrapper, data.shape, data.dtype)
        array = shared_data.get_data()
        array[:] = data[:]
        return shared_data



import uuid
from time import monotonic
from typing import Sequence, Set, Union, Optional

from .metrics.task import TaskMetrics
from .shm import SharedFieldsMixin


DEFAULT_PRIORITY = 0


class BaseTask(SharedFieldsMixin):
    task_id: Union[str, Sequence[str]] = ''
    priority: int = DEFAULT_PRIORITY
    expiration_time: Optional[float] = None
    metrics: Optional[TaskMetrics] = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.task_id = str(uuid.uuid4())
        self.priority = DEFAULT_PRIORITY
        self.metrics = TaskMetrics()

    def __repr__(self):
        name = self.__class__.__name__
        attrs = self.__dict__
        return f'<{name} {attrs}>'

    def __str__(self):
        return self.__repr__()

    def set_timeout(self, timeout_sec: float):
        self.expiration_time = monotonic() + timeout_sec

    def set_priority(self, priority: int):
        self.priority = priority

    def is_expired(self) -> bool:
        return self.expiration_time is not None and monotonic() >= self.expiration_time

    def _update_shared_fields(self, source_task: 'BaseTask') -> Set[str]:
        """Updates task shared fields and returns shared fields names.

        We shouldn't update existing shared fields and we use the overridden
        in SharedFieldsMixin `setattr` method to set SharedData values from source task.
        """
        for field, value in source_task._shared_fields.items():
            if field not in self._shared_fields:
                setattr(self, field, value)

        return {'_shared_fields'} | self._shared_fields.keys() | source_task._shared_fields.keys()

    def update(self, source_task: 'BaseTask', excluded_fields: Set[str] = None):
        """Sets data from source task.

        This is necessary in order to set the values from finished task to initial task
        in `process` Flow method.
        """
        immutable_fields = BaseTask.__dict__.keys()
        if excluded_fields:
            immutable_fields = immutable_fields | excluded_fields

        immutable_fields |= self._update_shared_fields(source_task)

        for field, value in source_task.__dict__.items():
            if field not in immutable_fields:
                setattr(self, field, value)


class StopTask(BaseTask):
    """Task for graceful shutdown."""



from setuptools import find_packages, setup

with open('README.rst') as f:
    long_description = f.read()

packages = ['aqueduct']

required_setup = [
    'cffi>=1.13.0,<2.0.0',
    'setuptools>=42.0.0,<81.0.0',
]

required = [
    'psutil==5.9.4',
]

extras = {
    'aiohttp': [
        'aiohttp',
    ],
    'numpy': [
        'numpy'
    ]
}

setup(
    name='aqueduct',
    packages=find_packages(),
    version='1.12.0',
    license='MIT',
    license_files='LICENSE.txt',
    author='Data Science SWAT',
    author_email='UnitDataScienceSwat@avito.ru',
    description='Builder for performance-efficient prediction.',
    url='https://github.com/avito-tech/aqueduct',
    download_url='https://github.com/avito-tech/aqueduct/archive/refs/heads/main.zip',
    keywords=['datascience', 'learning', ],
    python_requires='>=3.8',
    include_package_data=True,
    long_description=long_description,
    long_description_content_type='text/x-rst',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Topic :: Scientific/Engineering',
        'Topic :: Software Development :: Build Tools',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.8',
        'Operating System :: MacOS',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX',
        'Operating System :: Unix',
        'Operating System :: OS Independent',
    ],
    cffi_modules=["atomic_build.py:ffibuilder"],
    setup_requires=required_setup,
    install_requires=required,
    extras_require=extras,
)



"""Counts the execution time of each step of the service.

Steps: preprocessing of the image, processing with a model and postprocessing.

To start, see README.md.
"""
import os
import pathlib
import time
from typing import List

import logging
import numpy as np
import torch

from maas.maas.models import get_pipeline

IMAGES_DIR = pathlib.Path(__file__).parent / 'data'
pipeline = get_pipeline()


def _images(*im_paths: str):
    if not im_paths:
        im_paths = [os.path.join(IMAGES_DIR, name) for name in ('apple.jpg', 'cat.jpg', 'umbrella.jpg')]
    for im_path in im_paths:
        logging.info(f'Image: {im_path.split("/")[-1]}, size: {os.path.getsize(im_path)} bytes')
        with open(im_path, 'rb') as f:
            yield f.read()


def _get_times(func, *args, **kwargs) -> List[float]:
    """Returns time in seconds."""
    times = []
    for _ in range(20):
        t1 = time.time()
        func(*args, **kwargs)
        times.append((time.time() - t1))
    return times


def _mean_time(times: List[float], units: str = 'ms'):
    k = 1
    if units == 'ms':
        k = 10**3
    times = np.array(times) * k
    return f'{times.mean():.2f} +- {times.std():.2f} {units}'


def test_batch_processing():
    """Calculates just `process_list` method time.

    This time does not depend on image size because model uses image with fixed size.
    """
    logging.info('\nTest batch processing time\n')

    im = next(_images())
    im_torch = pipeline.preprocessor.process(im)

    def process(_im_torch, _batch_size):
        pipeline.classifier.process_list(data=[_im_torch] * _batch_size)
        torch.cuda.synchronize(0)

    for batch_size in [2 ** i for i in range(5)]:
        for _ in range(10):
            process(im_torch, batch_size)

        _times = [t / batch_size for t in _get_times(process, im_torch, batch_size)]
        logging.info(f'Batch size: {batch_size}\t Time: {_mean_time(_times)}')


def test_preprocessing():
    logging.info('\nTest preprocessing time\n')

    for im in _images():
        _times = _get_times(pipeline.preprocessor.process, im)
        logging.info(f'Time: {_mean_time(_times)}')


def test_postprocessing():
    logging.info('\nTest postprocessing time\n')

    im = next(_images())
    im_torch = pipeline.preprocessor.process(im)
    pred = pipeline.classifier.process_list(data=[im_torch])[0]

    _times = _get_times(pipeline.postprocessor.process, pred)
    logging.info(f'Time: {_mean_time(_times)}')


if __name__ == '__main__':
    test_preprocessing()
    test_batch_processing()
    test_postprocessing()



import time
from enum import Enum
from multiprocessing import Process, Queue
from typing import Any, Tuple

import numpy as np
import torch
from torch.multiprocessing import Process as t_Process, Queue as t_Queue

MAGIC_NUMBER = 42


class ImViewType(Enum):
    """Тип представления картинки."""
    NUMPY = 'numpy'
    TT = 'torch tensor'


class MPType(Enum):
    """Реализация мультипроцессинга."""
    PYTHON = 'python'
    TORCH = 'torch'


def get_mp_classes(mp_type: MPType) -> Tuple[type, type]:
    if mp_type == MPType.PYTHON:
        return Queue, Process
    if mp_type == MPType.TORCH:
        return t_Queue, t_Process


def get_image(im_size: Tuple[int, int] = (1280, 960)) -> np.ndarray:
    im = np.random.randint(0, 255, (*im_size, 3)).astype(np.float32)
    im[0][0][0] = MAGIC_NUMBER
    return im


def get_image_view(im: np.ndarray, view_type: ImViewType) -> Any:
    if view_type == ImViewType.NUMPY:
        return im
    if view_type == ImViewType.TT:
        return torch.from_numpy(im)


class StopProcess:
    pass


class timeit:  # noqa
    def __enter__(self):
        self.seconds = time.perf_counter()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.seconds = time.perf_counter() - self.seconds



from abc import ABC, abstractmethod
from typing import (
    Any,
    Optional,
    Protocol,
    Union,
)

from . import AQUEDUCT
from .collect import AqueductMetricsStorage


TRANSFER_TIME_PREFIX = 'transfer_time'
TASK_SIZE_PREFIX = 'task_size'
HANDLE_TIME_PREFIX = 'handle_time'
BATCH_TIME_PREFIX = 'batch_time'
BATCH_SIZE_PREFIX = 'batch_size'
QSIZE_PREFIX = 'qsize'
TASKS_PREFIX = 'tasks'
MEMORY_USAGE_PREFIX = 'memory_usage'
PROCESSES_PREFIX = 'processes'


class StatsDBuffer(Protocol):
    def count(self, name: str, value: Union[float, int]):
        pass

    def timing(self, name: str, value: Union[float, int]):
        pass

    def gauge(self, name: str, value: Union[float, int]):
        pass


class Exporter(ABC):
    def __init__(self, target: Any, export_period: float = 0.5, prefix: str = None):
        self.target = target
        self.export_period = export_period
        if prefix is None:
            self.prefix = AQUEDUCT
        else:
            self.prefix = f'{AQUEDUCT}.{prefix}'

    @abstractmethod
    def export(self, metrics: AqueductMetricsStorage):
        pass


class ToStatsDMetricsExporter(Exporter):
    """Translates aqueduct metrics to statsd metrics format."""
    def __init__(self, target: StatsDBuffer, export_period: float = 0.5, prefix: str = None):
        super().__init__(target, export_period, prefix)

    def export(self, metrics: AqueductMetricsStorage):
        for name, seconds in metrics.handle_times.items:
            self.target.timing(f'{self.prefix}.{HANDLE_TIME_PREFIX}.{name}', seconds * 1000)

        for name, seconds in metrics.transfer_times.items:
            self.target.timing(f'{self.prefix}.{TRANSFER_TIME_PREFIX}.{name}', seconds * 1000)

        for name, seconds in metrics.batch_times.items:
            self.target.timing(f'{self.prefix}.{BATCH_TIME_PREFIX}.{name}', seconds * 1000)

        for name, size in metrics.batch_sizes.items:
            self.target.timing(f'{self.prefix}.{BATCH_SIZE_PREFIX}.{name}', size)

        for name, size in metrics.queue_sizes.items:
            self.target.timing(f'{self.prefix}.{QSIZE_PREFIX}.{name}', size)

        for name, size in metrics.task_sizes.items:
            self.target.timing(f'{self.prefix}.{TASK_SIZE_PREFIX}.{name}', size)

        for name, cnt in metrics.tasks_stats.items:
            if cnt > 0:
                self.target.count(f'{self.prefix}.{TASKS_PREFIX}.{name}', cnt)

        for name, memory_usage in metrics.memory_usage.items:
            self.target.timing(f'{self.prefix}.{MEMORY_USAGE_PREFIX}.{name}', memory_usage)

        for name, cnt in metrics.processes_stats.items:
            if cnt > 0:
                self.target.count(f'{self.prefix}.{PROCESSES_PREFIX}.{name}', cnt)


class DummyExporter(Exporter):
    """Exports collected aqueduct metrics as is."""
    def __init__(self, target: Optional[AqueductMetricsStorage] = None, **kwargs):
        if target is None:
            target = AqueductMetricsStorage()
        super().__init__(target, **kwargs)

    def export(self, metrics: AqueductMetricsStorage):
        self.target.extend(metrics)



import subprocess


def guess_number(gte: int, lt: int):
    left, right = gte, lt
    number = left

    while True:
        val = (yield number)
        if val is not None:
            right = val
            break
        else:
            left = number
            number *= 2

    if number == gte:
        return

    while True:
        number = (left + right) // 2
        if number == left:
            break
        val = (yield number)
        if val is not None:
            right = val
        elif number > left:
            left = number

    yield number


class MaxNumberFinder:
    def __init__(self, gte: int = 1, lt: int = 10**3):
        self._numbers = guess_number(gte, lt)
        self.number = next(self._numbers)
        self.max_number = None

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type == StopIteration:
            return True

    def next(self, less: bool = False):
        if less:
            self.number = self._numbers.send(self.number)
        else:
            self.max_number = self.number
            self.number = next(self._numbers)

        if self.number == self.max_number:
            raise StopIteration


def run_cmd(command: str, stdout=None, shell=False, wait: bool = True) -> subprocess.Popen:
    cmd = command if shell else command.split()
    p = subprocess.Popen(cmd, stdout=stdout, shell=shell)
    if wait:
        p.wait()
    return p



import multiprocessing as mp
from multiprocessing.queues import Queue


class TaskMetricsQueue(Queue):
    def __init__(self, maxsize=0, *, ctx=None):
        ctx = ctx or mp.get_context()
        Queue.__init__(self, maxsize, ctx=ctx)
        self._task_size: int = 0

    def _recv_bytes_wrapper(self, *args, **kwargs):
        """Wraps _recv_bytes to record task byte size"""
        result = self.__recv_bytes(*args, **kwargs)
        self._task_size = len(result)
        return result

    @property
    def _recv_bytes(self):
        return self._recv_bytes_wrapper

    @_recv_bytes.setter
    def _recv_bytes(self, _recv_bytes_function):
        self.__recv_bytes = _recv_bytes_function

    @property
    def task_size(self) -> int:
        return self._task_size
